# Boltlight - a LN node wrapper
#
# Copyright (C) 2021-2022 boltlight contributors
# Copyright (C) 2018 inbitcoin s.r.l.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For a full list of contributors, please see the AUTHORS.md file.
"""Bitcoin and Lightning Network utils module."""

from decimal import Context, Decimal, Inexact, InvalidOperation
from logging import getLogger

from .. import boltlight_pb2 as pb
from .. import settings as sett
from ..errors import Err

LOGGER = getLogger(__name__)


# pylint: disable=too-many-arguments
def convert(context, source, target, amount, max_precision, enforce=None):
    """Convert amount from source to target unit.

    Round the result to the specified maximum precision. If the output unit
    matches the max precision, the value is returned as an integer.

    If enforce is set, check the value against the given boundaries.
    """
    if not amount:
        return 0
    try:
        amount = Decimal(str(amount))
    except InvalidOperation:
        # An invalid number would be blocked by the interface,
        # so this case has to come from the node
        return Err().internal_value_error(context, amount)
    if enforce:
        ratio = enforce['unit']['decimal'] - source['decimal']
        Enforcer.check_value(context, amount.scaleb(ratio), enforce)
    ratio = target['decimal'] - source['decimal']
    decimals = target['decimal'] - max_precision['decimal']
    try:
        converted = amount.scaleb(ratio)
        # cuts the amount to the required precision,
        # raising exceptions in case of inexact conversion
        result = converted.quantize(
            Decimal(1).scaleb(decimals),
            context=Context(traps=[Inexact, InvalidOperation]))
        if max_precision['decimal'] == target['decimal']:
            return int(result)
        return float(result)
    except (Inexact, InvalidOperation):
        return Err().value_error(context, amount)


# pylint: enable=too-many-arguments


def get_address_type(address):
    """Return the type of a bitcoin address."""
    if address[0] in ['b', 't']:
        return pb.Address.P2WPKH
    return pb.Address.NP2WPKH


def get_channel_balances(channels):
    """Calculate channel balances from a ListChannelsResponse."""
    out_tot = out_tot_now = in_tot = in_tot_now = 0
    for chan in channels:
        if chan.state != pb.Channel.OPEN:
            continue
        out_tot += chan.local_balance_msat
        in_tot += chan.remote_balance_msat
        if not chan.active:
            continue
        local_reserve = chan.local_reserve_sat * 1000
        remote_reserve = chan.remote_reserve_sat * 1000
        out_tot_now += max(0, chan.local_balance_msat - local_reserve)
        in_tot_now += max(0, chan.remote_balance_msat - remote_reserve)
    return pb.BalanceOffChainResponse(out_tot_msat=int(out_tot),
                                      out_tot_now_msat=int(out_tot_now),
                                      in_tot_msat=int(in_tot),
                                      in_tot_now_msat=int(in_tot_now))


def has_amount_encoded(payment_request):
    """Check if a bech32 payment request has any amount encoded."""
    separator = payment_request.rfind('1')
    hrp = payment_request[:separator]
    return _has_numbers(set(hrp))


def split_node_uri(ctx, node_uri):
    """Split a LN node URI into pubkey and host address.

    If the URI is invalid terminate the RPC with an invalid node_uri error.
    """
    try:
        pubkey, host_addr = node_uri.split('@')
        if not pubkey or not host_addr:
            raise ValueError
        return pubkey, host_addr
    except ValueError:
        return Err().invalid(ctx, 'node_uri')


def _has_numbers(input_string):
    """Check if string contains any number."""
    return any(char.isdigit() for char in input_string)


class Enforcer():  # pylint: disable=too-few-public-methods
    """Enforce BOLTs rules and value limits."""

    BTC = {'name': 'btc', 'decimal': 0}
    MBTC = {'name': 'mbtc', 'decimal': 3}
    BITS = {'name': 'bits', 'decimal': 6}
    SATS = {'name': 'sats', 'decimal': 8}
    MSATS = {'name': 'msats', 'decimal': 11}

    DEFAULT = {'min_value': 0, 'unit': MSATS}

    # BOLT2 suggests (but does not enforce) a reserve of 1% of the channel
    # funding total, and the reserve cannot be lower than the dust limit.
    FUNDING_SATOSHIS = {
        'min_value': 100 * sett.DUST_LIMIT_SAT,
        'max_value': 2**24,
        'unit': SATS
    }
    PUSH_MSAT = {'max_value': 2**24 * 1000, 'unit': MSATS}
    LN_PAYREQ = {'min_value': 0, 'max_value': 2**32, 'unit': MSATS}
    LN_TX = {'min_value': 1, 'max_value': 2**32, 'unit': MSATS}

    OC_TX = {'min_value': 1, 'max_value': 21 * 10**14, 'unit': SATS}
    # assuming minimum 220 bytes tx and single 21M BTC input
    OC_FEE = {'min_value': 1, 'max_value': 21 * 10**14 // 220, 'unit': SATS}

    MIN_FINAL_CLTV_EXPIRY = {'min_value': 1, 'max_value': 5 * 10**8}
    CLTV_EXPIRY_DELTA = {'min_value': 1, 'max_value': 2**16}

    @staticmethod
    def check_value(context, value, enforce=None):
        """Check that value is between min_value and max_value."""
        if not enforce:
            enforce = Enforcer.DEFAULT
        if sett.ENFORCE:
            if 'min_value' in enforce and value < enforce['min_value']:
                Err().value_too_low(context, value)
            if 'max_value' in enforce and value > enforce['max_value']:
                Err().value_too_high(context, value)
        return True



# Boltlight - a LN node wrapper
#
# Copyright (C) 2021-2022 boltlight contributors
# Copyright (C) 2018 inbitcoin s.r.l.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Security utils module."""

from marshal import dumps as mdumps
from marshal import loads as mloads

from .. import boltlight_pb2 as pb
from .. import settings as sett
from ..errors import Err
from .db import get_secret_from_db, get_token_from_db, session_scope
from .network import check_req_params


def check_password(context, session, password):
    """Check if the inserted password is correct.

    To perform the check, generate an access key and try to decrypt the token
    in the DB.
    """
    encrypted_token, params = get_token_from_db(session)
    token_params = ScryptParams('')
    token_params.deserialize(params)
    derived_key = Crypter.gen_derived_key(password, token_params)
    clear_token = Crypter.decrypt(context, encrypted_token, derived_key)
    if clear_token != sett.ACCESS_TOKEN:
        Err().wrong_password(context)


# pylint: disable=too-many-arguments
def get_secret(context, session, password, impl, sec_type, active_only=False):
    """Retrieve and decrypt an implementation secret from the DB."""
    impl_secret = get_secret_from_db(session, impl, sec_type)
    if not impl_secret or not impl_secret.secret:
        return None
    if active_only and not impl_secret.active:
        return None
    params = ScryptParams('')
    params.deserialize(impl_secret.scrypt_params)
    derived_key = Crypter.gen_derived_key(password, params)
    return Crypter.decrypt(context, impl_secret.secret, derived_key)


# pylint: enable=too-many-arguments


def unlock_node_with_password(context, req, unlock_method):
    """Check password in request and call the given unlock method."""
    check_req_params(context, req, 'password')
    with session_scope(context) as session:
        check_password(context, session, req.password)
        unlock_method(context, req.password, session=session)
    return pb.UnlockNodeResponse()


class Crypter():
    """Provide cryptographic methods."""
    @staticmethod
    def gen_derived_key(password, scrypt_params):
        """Derive a key from a password using Scrypt."""
        # pylint: disable=import-outside-toplevel
        from pylibscrypt import scrypt

        # pylint: enable=import-outside-toplevel
        return scrypt(bytes(password, 'utf-8'),
                      scrypt_params.salt,
                      N=scrypt_params.cost_factor,
                      r=scrypt_params.block_size_factor,
                      p=scrypt_params.parallelization_factor,
                      olen=scrypt_params.key_len)

    @staticmethod
    def crypt(clear_data, derived_key):
        """Encrypt data using Secretbox and the access key.

        Return the encrypted data in a serialized form.
        """
        # pylint: disable=import-outside-toplevel
        from nacl.secret import SecretBox

        # pylint: enable=import-outside-toplevel
        return SecretBox(derived_key).encrypt(clear_data)

    @staticmethod
    def decrypt(context, encrypted_data, derived_key):
        """Decrypt serialized data using Secretbox and the access key.

        Terminate the RPC when password is wrong.
        """
        # pylint: disable=import-outside-toplevel
        from nacl.exceptions import CryptoError
        from nacl.secret import SecretBox

        # pylint: enable=import-outside-toplevel
        try:
            return SecretBox(derived_key).decrypt(encrypted_data)
        except CryptoError:
            return Err().wrong_password(context)


class ScryptParams():
    """Convenient class to store scrypt parameters."""

    # pylint: disable=too-many-arguments
    def __init__(self,
                 salt,
                 cost_factor=sett.SCRYPT_PARAMS['cost_factor'],
                 block_size_factor=sett.SCRYPT_PARAMS['block_size_factor'],
                 parallelization_factor=sett.
                 SCRYPT_PARAMS['parallelization_factor'],
                 key_len=sett.SCRYPT_PARAMS['key_len']):
        self.salt = salt
        self.cost_factor = cost_factor
        self.block_size_factor = block_size_factor
        self.parallelization_factor = parallelization_factor
        self.key_len = key_len

    # pylint: enable=too-many-arguments

    def serialize(self):
        """Serialize ScryptParams."""
        return mdumps([
            self.salt, self.cost_factor, self.block_size_factor,
            self.parallelization_factor, self.key_len
        ])

    def deserialize(self, serialized):
        """Deserialize ScryptParams."""
        deserialized = mloads(serialized)
        self.salt = deserialized[0]
        self.cost_factor = deserialized[1]
        self.block_size_factor = deserialized[2]
        self.parallelization_factor = deserialized[3]
        self.key_len = deserialized[4]



# Boltlight - a LN node wrapper
#
# Copyright (C) 2021-2022 boltlight contributors
# Copyright (C) 2018 inbitcoin s.r.l.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Database utils module."""

from contextlib import contextmanager
from logging import getLogger
from os import path
from pathlib import Path
from platform import system

from alembic.command import stamp
from alembic.config import Config
from alembic.runtime import migration
from alembic.script import ScriptDirectory
from sqlalchemy import create_engine
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm import sessionmaker

from .. import settings as sett
from ..db import AccessToken, Base, ImplementationSecret, MacRootKey
from ..errors import Err

LOGGER = getLogger(__name__)

ENGINE = None
Session = None


@contextmanager
def session_scope(context):
    """Provide a transactional scope around a series of operations."""
    session = Session()
    try:
        yield session
        session.commit()
    except SQLAlchemyError:
        session.rollback()
        Err().db_error(context)
    except Exception as exc:
        session.rollback()
        raise exc
    finally:
        session.close()


def detect_impl_secret(session):
    """Detect if implementation has a secret stored."""
    if sett.IMPLEMENTATION == 'clightning':
        return False
    detected = False
    error = False
    impl_secret = get_secret_from_db(session, sett.IMPLEMENTATION,
                                     sett.IMPL_SEC_TYPE)
    if sett.IMPLEMENTATION in ('eclair', 'electrum'):
        detected = True  # secret always necessary when using eclair/electrum
        if not impl_secret or not impl_secret.secret:
            error = True
    if sett.IMPLEMENTATION == 'lnd':
        if impl_secret and impl_secret.active:
            detected = True
            if not impl_secret.secret:
                error = True
    if error:
        raise RuntimeError(
            'Cannot obtain implementation secret, add it by running '
            'boltlight-secure')
    return detected


def init_db(new_db=False, alembic_cfg=None):
    """Initialize DB connection, creating missing tables if requested."""
    if not alembic_cfg:
        alembic_cfg = get_alembic_cfg(new_db)
    sec = 'boltlight_log'
    alembic_cfg.set_section_option(
        sec, 'boltlight', path.join(sett.LOGS_DIR, sett.LOGS_BOLTLIGHT))
    alembic_cfg.set_section_option(
        sec, 'migrations', path.join(sett.LOGS_DIR, sett.LOGS_MIGRATIONS))
    global ENGINE  # pylint: disable=global-statement
    global Session  # pylint: disable=global-statement
    ENGINE = create_engine(_get_db_url(new_db))
    Session = sessionmaker(bind=ENGINE, autoflush=False, autocommit=False)
    if new_db:
        LOGGER.info('Creating database')
        Base.metadata.create_all(ENGINE)
        stamp(alembic_cfg, 'head')


def get_alembic_cfg(new_db):
    """Return alembic Config object."""
    alembic_cfg = Config(sett.ALEMBIC_CFG)
    alembic_cfg.set_main_option('sqlalchemy.url', _get_db_url(new_db))
    alembic_cfg.set_main_option('script_location',
                                sett.PKG_NAME + ':migrations')
    return alembic_cfg


def _get_db_url(new_db):
    """Construct the DB's URL for SQLAlchemy.

    Fail when DB is missing at runtime.
    """
    db_abspath = ''
    db_relpath = Path(sett.DB_DIR).joinpath(sett.DB_NAME)
    try:
        db_abspath = db_relpath.resolve(strict=True)
    except FileNotFoundError:
        if new_db:
            db_relpath.touch()
            db_abspath = db_relpath.resolve()
        else:
            raise RuntimeError('Your database is missing. Create it by '
                               'running boltlight-secure') from None
    running_sys = system()
    if running_sys in ('Linux', 'Darwin'):
        return f'sqlite:///{db_abspath}'
    if running_sys == 'Windows':
        return rf'sqlite:///{db_abspath}'
    LOGGER.warning('Unrecognized OS, using in-memory database')
    return 'sqlite://'


def is_db_ok(session, configuring=False):
    """Return whether the DB is ok.

    DB is ok when it doesn't contain old data and it's not missing essential
    data.
    """
    # checking if old salt table exists
    if ENGINE.dialect.has_table(ENGINE, 'salt_table'):
        return False
    # checking if encrypted token exists
    if not ENGINE.dialect.has_table(ENGINE, AccessToken.__tablename__) or \
            not get_token_from_db(session):
        return False
    # checking if macaroon root key exists
    if not configuring and not sett.DISABLE_MACAROONS:
        if not ENGINE.dialect.has_table(ENGINE, MacRootKey.__tablename__) or \
                not get_mac_params_from_db(session):
            LOGGER.error('Please make sure you have generated macaroon at '
                         'least one time')
            return False
    # checking if implementation_secrets table exists
    if not ENGINE.dialect.has_table(ENGINE, 'implementation_secrets'):
        return False
    if not _is_db_at_head(get_alembic_cfg(False), ENGINE):
        LOGGER.error('Migrations may not have applied correctly')
        return False
    return True


def _is_db_at_head(alembic_cfg, connectable):
    """Return whether the DB revision is at head."""
    directory = ScriptDirectory.from_config(alembic_cfg)
    with connectable.begin() as connection:
        getLogger('alembic').propagate = False
        context = migration.MigrationContext.configure(connection)
        return set(context.get_current_heads()) == set(directory.get_heads())


def get_mac_params_from_db(session):
    """Get macaroon root key parameters from database."""
    mac_params = session.query(MacRootKey).first()
    if not mac_params:
        return None
    return mac_params.scrypt_params


def get_secret_from_db(session, implementation, sec_type):
    """Get implementation's secret from database."""
    sec = session.query(ImplementationSecret).filter_by(
        implementation=implementation, secret_type=sec_type).first()
    return sec


def get_token_from_db(session):
    """Get the encrypted token from database."""
    access_token = session.query(AccessToken).first()
    if not access_token:
        return None, None
    return access_token.data, access_token.scrypt_params


def save_mac_params_to_db(session, scrypt_params):
    """Save macaroon root key parameters in database."""
    session.merge(MacRootKey(data='mac_params', scrypt_params=scrypt_params))


# pylint: disable=too-many-arguments
def save_secret_to_db(session, implementation, sec_type, active, data,
                      scrypt_params):
    """Save implementation's secret in database."""
    session.merge(
        ImplementationSecret(implementation=implementation,
                             secret_type=sec_type,
                             active=active,
                             secret=data,
                             scrypt_params=scrypt_params))


# pylint: enable=too-many-arguments


def save_token_to_db(session, token, scrypt_params):
    """Save the encrypted token in database."""
    session.merge(AccessToken(data=token, scrypt_params=scrypt_params))



# Copyright (C) 2018 inbitcoin s.r.l.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""The errors module for boltlight."""

from importlib import import_module
from logging import getLogger
from re import sub

from grpc import StatusCode

from . import settings as sett

LOGGER = getLogger(__name__)

ERRORS = {
    'amount_required': {
        'code': 'INVALID_ARGUMENT',
        'msg': 'A positive amount is required for empty invoices'
    },
    'connect_failed': {
        'code': 'CANCELLED',
        'msg': 'Connection to peer failed'
    },
    'closechannel_failed': {
        'code': 'CANCELLED',
        'msg': 'Failed to close channel'
    },
    'db_error': {
        'code': 'CANCELLED',
        'msg': 'Error accessing database'
    },
    'incorrect_description': {
        'code': 'INVALID_ARGUMENT',
        'msg': 'Provided description doesn\'t match the payment request one'
    },
    'insufficient_fee': {
        'code': 'OUT_OF_RANGE',
        'msg': 'Fees are insufficient'
    },
    'insufficient_funds': {
        'code': 'FAILED_PRECONDITION',
        'msg': 'Funds are insufficient'
    },
    'invalid': {
        'code': 'INVALID_ARGUMENT',
        'msg': "Invalid parameter '%PARAM%'"
    },
    'invoice_expired': {
        'code': 'OUT_OF_RANGE',
        'msg': 'Invoice expired'
    },
    'invoice_not_found': {
        'code': 'NOT_FOUND',
        'msg': 'Invoice not found'
    },
    'missing_parameter': {
        'code': 'INVALID_ARGUMENT',
        'msg': "Parameter '%PARAM%' is necessary"
    },
    'node_error': {
        'code': 'UNAVAILABLE',
        'msg': '[node error] %PARAM%'
    },
    'node_locked': {
        'code': 'UNAVAILABLE',
        'msg': 'Node locked, unlock it by calling UnlockNode'
    },
    'openchannel_failed': {
        'code': 'CANCELLED',
        'msg': 'Failed to open channel'
    },
    'payinvoice_failed': {
        'code': 'CANCELLED',
        'msg': 'Invoice payment has failed'
    },
    'payonchain_failed': {
        'code': 'CANCELLED',
        'msg': 'On-chain payment has failed'
    },
    'route_not_found': {
        'code': 'NOT_FOUND',
        'msg': 'Can\'t find route to node'
    },
    'unimplemented_method': {
        'code':
        'UNIMPLEMENTED',
        'msg': ("The gRPC method '%PARAM%' is not supported for this "
                'implementation')
    },
    'unimplemented_parameter': {
        'code':
        'UNIMPLEMENTED',
        'msg': ("The gRPC parameter '%PARAM%' is not supported for this "
                'implementation')
    },
    'unimplemented_param_value': {
        'code':
        'UNIMPLEMENTED',
        'msg': ("Field '%PARAM%' doesn't support value '%PARAM%' on this "
                "implementation")
    },
    'unsettable': {
        'code': 'INVALID_ARGUMENT',
        'msg': "Parameter '%PARAM%' unsettable"
    },
    'value_error': {
        'code': 'INVALID_ARGUMENT',
        'msg': "Value '%PARAM%' exceeds maximum precision"
    },
    'value_too_low': {
        'code': 'OUT_OF_RANGE',
        'msg': "Value '%PARAM%' is below minimum threshold"
    },
    'value_too_high': {
        'code': 'OUT_OF_RANGE',
        'msg': "Value '%PARAM%' exceeds maximum treshold"
    },
    'wrong_node_password': {
        'code':
        'INVALID_ARGUMENT',
        'msg': ('Stored node password is incorrect, update it by running '
                'boltlight-secure')
    },
    'wrong_password': {
        'code': 'INVALID_ARGUMENT',
        'msg': 'Wrong password'
    },
    'internal_value_error': {
        'code': 'INTERNAL',
        'msg': "Value '%PARAM%' received from node is not a number"
    },
    # Fallback
    'unexpected_error': {
        'code': 'UNKNOWN'
    }
}


class Err():  # pylint: disable=too-few-public-methods
    """Boltlight errors class."""
    def __getattr__(self, name):
        """Dispatch the called error dynamically."""
        def error_dispatcher(context, *params):
            """Abort gRPC call raising a grpc.RpcError to the API caller."""
            if name in ERRORS:
                scode = getattr(StatusCode, ERRORS[name]['code'])
                msg = ''
                if 'msg' in ERRORS[name]:
                    msg = ERRORS[name]['msg']
                if params:
                    for param in params:
                        msg = sub('%PARAM%', str(param), msg, 1)
                if name == 'unexpected_error':
                    msg = params[0]
                    LOGGER.error('Unexpected error: %s', msg)
                LOGGER.error('> %s', msg)
                context.abort(scode, msg)
            else:
                LOGGER.error('Unmapped error key')

        return error_dispatcher

    def report_error(self, context, error, always_abort=True):
        """Call the proper error method or throw an unexpected_error."""
        module = import_module(f'..light_{sett.IMPLEMENTATION}', __name__)
        for msg, act in module.ERRORS.items():
            if msg in error:
                args = [context, act['params']] if 'params' in act \
                    else [context]
                if act['fun'] == 'node_error':
                    args = [context, error]
                getattr(self, act['fun'])(*args)
        if always_abort:
            self.unexpected_error(context, str(error))



# Boltlight - a LN node wrapper
#
# Copyright (C) 2021-2022 boltlight contributors
# Copyright (C) 2018 inbitcoin s.r.l.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Security operations module.

Handle implementation secrets, macaroons creation and storage.
"""

import sys
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import TimeoutError as TimeoutErrFut
from configparser import Error as ConfigError
from contextlib import suppress
from datetime import datetime, timedelta, timezone
from getpass import getpass
from logging import CRITICAL, getLogger
from os import devnull, environ, path, remove, urandom
from select import select
from signal import SIGTERM, signal
from time import sleep, time

from sqlalchemy.exc import SQLAlchemyError

from . import settings as sett
from .macaroons import MAC_VERSION, MACAROONS, get_baker
from .utils.db import (
    init_db, is_db_ok, save_mac_params_to_db, save_secret_to_db,
    save_token_to_db, session_scope)
from .utils.exceptions import InterruptException
from .utils.misc import (
    die, handle_importerror, handle_keyboardinterrupt, handle_sigterm,
    init_common, str2bool)
from .utils.network import FakeContext
from .utils.security import Crypter, ScryptParams, check_password, get_secret

signal(SIGTERM, handle_sigterm)

LOGGER = getLogger(__name__)


def _consume_bytes(consumable, num):
    """Consume num elements from a byte string and return them."""
    consumable = list(consumable)
    i = 0
    data = []
    while i < num:
        data.append(consumable.pop(0))
        i = i + 1
    return bytes(data)


def _remove_files():
    """Remove database and macaroons."""
    LOGGER.info('Removing database and macaroons')
    with suppress(FileNotFoundError):
        remove(sett.DB_PATH)
        for file_name in MACAROONS:
            macaroon_file = path.join(sett.MACAROONS_DIR, file_name)
            remove(macaroon_file)


def _get_eclair_password():
    """Get eclair's password from stdin."""
    while True:
        data = getpass('Insert eclair password (cannot be empty): ')
        if data:
            more_data = getpass('Insert eclair password again: ')
            if data != more_data:
                die('Passwords do not match')
            return data.encode()


def _get_electrum_password():
    """Get electrum's password from stdin."""
    while True:
        data = getpass('Insert electrum password (cannot be empty): ')
        if data:
            more_data = getpass('Insert electrum password again: ')
            if data != more_data:
                die('Passwords do not match')
            return data.encode()


def _get_lnd_macaroon(mac_path=None):
    """Get lnd's macaroon from file."""
    def _read_macaroon(mac_path):
        print('Reading lnd macaroon from the provided path...')
        try:
            with open(mac_path, 'rb') as file:
                return file.read()
        except OSError as err:
            return die('Cannot read macaroon file: ' + str(err))

    if mac_path:
        return _read_macaroon(mac_path)
    mac_path = input('If your lnd instance requires a macaroon for '
                     'authorization, provide its path\nhere (filename included'
                     ', overrides current one if any) or just press enter to\n'
                     'provide none (skip) ')
    if mac_path:
        return _read_macaroon(mac_path)
    print('You have not provided a path, usage of lnd without macaroon '
          '(insecure)')
    return None


def _get_lnd_password():
    """Get lnd's password from stdin."""
    data = getpass('Insert lnd password or press enter to skip '
                   '(node unlocking will not be available): ')
    if data:
        more_data = getpass('Insert lnd password again: ')
        if data != more_data:
            die('Passwords do not match')
        return data.encode()
    return None


def _set_eclair_password(secret):
    """Handle storage of eclair's password."""
    data = None
    activate_secret = 1
    if not secret:
        data = _get_eclair_password()
    else:
        data = secret
        rm_sec = input("A password for eclair is already stored, "
                       "do you want to update it? [y/N] ")
        if str2bool(rm_sec):
            data = _get_eclair_password()
    return data, activate_secret, 'password'


def _set_electrum_password(secret):
    """Handle storage of electrum's password."""
    data = None
    activate_secret = 1
    if not secret:
        data = _get_electrum_password()
    else:
        data = secret
        rm_sec = input("A password for electrum is already stored, "
                       "do you want to update it? [y/N] ")
        if str2bool(rm_sec):
            data = _get_electrum_password()
    return data, activate_secret, 'password'


def _set_lnd_macaroon(secret):
    """Handle storage of lnd's macaroon."""
    data = None
    activate_secret = 1
    if not secret:
        data = _get_lnd_macaroon()
    else:
        data = secret
        rm_sec = input("A macaroon for lnd is already stored, "
                       "do you want to update it? [y/N] ")
        if str2bool(rm_sec):
            data = _get_lnd_macaroon()
    if data:
        res = input("Connect to lnd using its macaroon "
                    "(warning: insecure without)? [Y/n] ")
        if not str2bool(res, force_true=True):
            activate_secret = 0
    else:
        activate_secret = 0
    return data, activate_secret, 'macaroon'


def _set_lnd_password(secret):
    """Handle storage of lnd's password."""
    data = None
    activate_secret = 0
    if not secret:
        data = _get_lnd_password()
    else:
        data = secret
        rm_sec = input("A password for lnd is already stored, "
                       "do you want to update it? [y/N] ")
        if str2bool(rm_sec):
            data = _get_lnd_password()
    if data:
        activate_secret = 1
    return data, activate_secret, 'password'


def _save_token(session, password, scrypt_params):
    """Encrypt token and save it into the DB to verify password correctness."""
    derived_key = Crypter.gen_derived_key(password, scrypt_params)
    encrypted_token = Crypter.crypt(sett.ACCESS_TOKEN, derived_key)
    save_token_to_db(session, encrypted_token, scrypt_params.serialize())
    LOGGER.info('Encrypted token stored in the DB')


# pylint: disable=too-many-arguments
def _save_secret(session,
                 password,
                 scrypt_params,
                 sec_type,
                 secret,
                 activate_secret,
                 implementation=None):
    """Encrypt implementation secret and save it into the DB."""
    encrypted_secret = None
    if not implementation:
        implementation = sett.IMPLEMENTATION
    if secret:
        derived_key = Crypter.gen_derived_key(password, scrypt_params)
        encrypted_secret = Crypter.crypt(secret, derived_key)
    save_secret_to_db(session, implementation, sec_type, activate_secret,
                      encrypted_secret, scrypt_params.serialize())


# pylint: enable=too-many-arguments


def _recover_secrets(session, password):
    """Recover secrets from DB, making sure password is correct."""
    ecl_pass = lnd_mac = lnd_pass = None
    try:
        ecl_pass = get_secret(FakeContext(), session, password, 'eclair',
                              'password')
        ele_pass = get_secret(FakeContext(), session, password, 'electrum',
                              'password')
        lnd_mac = get_secret(FakeContext(), session, password, 'lnd',
                             'macaroon')
        lnd_pass = get_secret(FakeContext(), session, password, 'lnd',
                              'password')
    except RuntimeError as err:
        die(err)
    return ecl_pass, ele_pass, lnd_mac, lnd_pass


def _create_macaroons(session, password, scrypt_params):
    """Create the macaroon files."""
    print('Creating macaroons...')
    sett.MAC_ROOT_KEY = Crypter.gen_derived_key(password, scrypt_params)
    save_mac_params_to_db(session, scrypt_params.serialize())
    baker = get_baker(sett.MAC_ROOT_KEY)
    for file_name, permitted_ops in MACAROONS.items():
        macaroon_file = path.join(sett.MACAROONS_DIR, file_name)
        expiration_time = datetime.now(tz=timezone.utc) + timedelta(days=365)
        caveats = None
        mac = baker.oven.macaroon(MAC_VERSION, expiration_time, caveats,
                                  permitted_ops)
        serialized_macaroon = mac.macaroon.serialize()
        with open(macaroon_file, 'wb') as file:
            file.write(serialized_macaroon.encode())
        LOGGER.info('%s written to %s', file_name, sett.MACAROONS_DIR)


def _get_entropy():
    """Get the available entropy."""
    with open('/proc/sys/kernel/random/entropy_avail', 'r',
              encoding='utf-8') as entropy:
        entropy_available = int(entropy.read())
        return entropy_available


def _input():
    """Get an input asynchronously."""
    read_list = [sys.stdin]
    sett.FIRST_WORK_TIME = time()
    while read_list and sett.COLLECTING_INPUT:
        ready = select(read_list, [], [], 0.2)[0]
        if not ready and sett.COLLECTING_INPUT:
            _idle()
        else:
            for file in ready:
                line = file.readline()
                if not line:  # EOF, remove file from input list
                    read_list.remove(file)
                elif line.rstrip():
                    return line.rstrip()


def _idle():
    """Print periodic messages during input idling."""
    idle_msg = sett.IDLE_MESSAGES[sett.IDLE_COUNTER]
    if time() - sett.FIRST_WORK_TIME > idle_msg['delay']:
        print(idle_msg['msg'])
        if sett.IDLE_COUNTER == 3:
            idle_msg['delay'] = idle_msg['delay'] + 15
        else:
            sett.IDLE_COUNTER += 1


def _read(source, num_bytes):
    """Get entropy from /dev/random asynchronously."""
    try:
        while _get_entropy() < num_bytes * 8 * 1.2:
            if not sett.SEARCHING_ENTROPY:
                return None
            sleep(1)
        random_bytes = source.read(num_bytes)
        return random_bytes
    except IOError:
        if not sett.SEARCHING_ENTROPY:
            return None
        use_urand = input("Cannot retrieve available entropy, do you want to "
                          "use whatever your os provides\nto python's "
                          "os.urandom? [Y/n] ")
        if str2bool(use_urand, force_true=True):
            return None
        return die("No way to retrieve the amount of available entropy")


def _gen_random_data(num_bytes):  # pylint: disable=too-many-branches
    """Generate random data of key_len length."""
    unsafe_secrets = environ.get('unsafe_secrets', 0)
    if unsafe_secrets:
        return urandom(num_bytes)
    print('Trying to collect entropy...')
    future_input = None
    try:
        with open("/dev/random", "rb") as source:
            with ThreadPoolExecutor(max_workers=2) as executor:
                future_read = executor.submit(_read, source, num_bytes)
                data = future_read.result(timeout=1)
                if data:
                    return data
    except FileNotFoundError:
        use_urand = input("The blocking '/dev/random' entropy source is "
                          "not available, do you want to use\nwhatever "
                          "your os provides to python's os.urandom? "
                          "[Y/n] ")
        if str2bool(use_urand, force_true=True):
            return urandom(num_bytes)
        die("No Random Numbers Generator available")
    except TimeoutErrFut:
        print("This call might take long depending on the "
              "available entropy.\nIf this happens you can:\n"
              " - type randomly on the keyboard or move the mouse\n"
              " - install entropy collecting tools like haveged\n"
              " - install a hardware TRNG\n"
              " - type 'unsafe' and press enter to use a non-blocking "
              "entropy source; this choice will NOT be remembered for "
              "later\n")
        while future_read.running():
            if not future_input:
                future_input = executor.submit(_input)
            if future_input.done():
                if future_input.result() == 'unsafe':
                    sett.SEARCHING_ENTROPY = False
                    sett.COLLECTING_INPUT = False
                    print('Unsafe mode selected')
                    return urandom(num_bytes)
                future_input = None
            sleep(1)
        sett.COLLECTING_INPUT = False
        print("\nEnough entropy was collected, thanks for waiting")
        result = future_read.result()
        if result:
            return result
    finally:
        if source:
            source.close()
        if executor:
            executor.shutdown(wait=False)
    # should never get here
    return urandom(num_bytes)


def _gen_password(seed):
    """Generate a safe random password from a 64-char alphabet."""
    # base58 charset plus some symbols up to 64
    alpha = r'123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz+/\&-_'
    assert 256 % len(alpha) == 0
    return ''.join(alpha[i % len(alpha)] for i in seed)


def _get_req_salt_len(new, interactive=True):
    """Determine the maximum bytes of salt that will be needed."""
    secrets = 1  # for boltlight's macaroons
    if sett.IMPLEMENTATION == 'eclair':
        secrets += 1  # for eclair password
    elif sett.IMPLEMENTATION == 'lnd':
        secrets += 2  # for lnd password and macaroon
    for_password = 0
    if new:
        secrets += 1  # for access token
        if interactive:
            for_password = sett.PASSWORD_LEN  # for auto-generated password
    return sett.SALT_LEN * secrets + for_password


def _db_exists():
    """Check if DB exists."""
    return path.exists(sett.DB_PATH)


def _keep_db(interactive):
    """Decide whether to keep the DB or delete it.

    In interactive mode, ask the user if he wants to delete the DB.
    Return True if the DB has been kept, False otherwise.
    """
    if not interactive:
        return True
    delete = input('Db already exists, do you want to override it? (note '
                   'this will also delete\nmacaroon files) [y/N] ')
    if str2bool(delete):
        _remove_files()
        return False
    return True


def db_config_interactive(session, new):  # pylint: disable=too-many-branches
    """Configure a new or existing database interactively."""
    ecl_pass = ele_pass = lnd_mac = lnd_pass = None
    if new:
        print('Boltlight is about to ask for a new password! As humans are '
              'really bad at\ngenerating entropy, we suggest using a password '
              'manager to generate and store\nthe password on your behalf. '
              'Refer to doc/security.md for more details.')
        gen_psw = input('Do you want boltlight to generate a safe random '
                        'password for you? (new password\nwill be printed to '
                        'stdout) [Y/n] ')
        salts_len = _get_req_salt_len(new)
        if str2bool(gen_psw, force_true=True):
            seed = _gen_random_data(sett.PASSWORD_LEN + salts_len)
            password = _gen_password(_consume_bytes(seed, sett.PASSWORD_LEN))
            print("Here is your new password:")
            print(password)
        else:
            seed = _gen_random_data(salts_len)
            password = getpass('Insert a safe password for boltlight: ')
        password_check = getpass('Save the password and then enter it '
                                 'for verification: ')
        if password != password_check:
            _remove_files()
            die("Passwords do not match")
        scrypt_params = ScryptParams(_consume_bytes(seed, sett.SALT_LEN))
        _save_token(session, password, scrypt_params)
    else:
        if not is_db_ok(session, configuring=True):
            die('Detected an incomplete configuration. Delete database.')
        password = getpass("Insert boltlight's password: ")
        try:
            check_password(FakeContext(), session, password)
        except RuntimeError:
            die('Wrong password')
        ecl_pass, ele_pass, lnd_mac, lnd_pass = _recover_secrets(
            session, password)
        seed = _gen_random_data(_get_req_salt_len(new))
    create_mac = input('Do you want to create macaroons (warning: generated '
                       'files should not be kept in\nthis host)? [y/N] ')
    if str2bool(create_mac):
        scrypt_params = ScryptParams(_consume_bytes(seed, sett.SALT_LEN))
        _create_macaroons(session, password, scrypt_params)
    secrets = []
    if sett.IMPLEMENTATION == 'eclair':
        secrets = [_set_eclair_password(ecl_pass)]
    if sett.IMPLEMENTATION == 'electrum':
        secrets = [_set_electrum_password(ele_pass)]
    if sett.IMPLEMENTATION == 'lnd':
        secrets = [_set_lnd_macaroon(lnd_mac), _set_lnd_password(lnd_pass)]
    if secrets:  # user gave us secrets to encrypt and save
        for secret in secrets:
            scrypt_params = ScryptParams(_consume_bytes(seed, sett.SALT_LEN))
            _save_secret(session, password, scrypt_params, secret[2],
                         secret[0], secret[1])


def db_config_non_interactive(session, new, password):
    """Configure a new or existing database in batch-mode."""
    ecl_pass = ele_pass = lnd_mac = lnd_pass = None
    if new:
        salts_len = _get_req_salt_len(new, interactive=False)
        seed = _gen_random_data(salts_len)
        scrypt_params = ScryptParams(_consume_bytes(seed, sett.SALT_LEN))
        _save_token(session, password, scrypt_params)
    else:
        if not is_db_ok(session):
            die('Detected an incomplete configuration. Delete database.')
        try:
            check_password(FakeContext(), session, password)
        except RuntimeError:
            die('Wrong password')
        ecl_pass, ele_pass, lnd_mac, lnd_pass = \
            _recover_secrets(session, password)
        seed = _gen_random_data(_get_req_salt_len(new, interactive=False))
    create_mac = environ.get('create_macaroons', 0)
    if create_mac:
        scrypt_params = ScryptParams(_consume_bytes(seed, sett.SALT_LEN))
        _create_macaroons(session, password, scrypt_params)
    secrets = []
    ecl_pass = environ.get('eclair_password')
    ele_pass = environ.get('electrum_password')
    lnd_mac = environ.get('lnd_macaroon')
    lnd_pass = environ.get('lnd_password')
    if ecl_pass:
        secrets.append([ecl_pass.encode(), 1, 'password', 'eclair'])
    if ele_pass:
        secrets.append([ele_pass.encode(), 1, 'password', 'electrum'])
    if lnd_mac:
        secrets.append([_get_lnd_macaroon(lnd_mac), 1, 'macaroon', 'lnd'])
    if lnd_pass:
        secrets.append([lnd_pass.encode(), 1, 'password', 'lnd'])
    if secrets:  # user gave us secrets to encrypt and save
        for secret in secrets:
            scrypt_params = ScryptParams(_consume_bytes(seed, sett.SALT_LEN))
            _save_secret(session,
                         password,
                         scrypt_params,
                         secret[2],
                         secret[0],
                         secret[1],
                         implementation=secret[3])


def secure():
    """Secure entrypoint."""
    try:
        _secure()
    except ImportError as err:
        handle_importerror(err)
    except RuntimeError as err:
        die(str(err))
    except ConfigError as err:
        err_msg = ''
        if str(err):
            err_msg = str(err)
        die('Configuration error: ' + err_msg)
    except SQLAlchemyError as err:
        err_msg = ''
        if str(err):
            err_msg = str(err)
        die('DB error: ' + err_msg)
    except InterruptException:
        sett.COLLECTING_INPUT = False
        sett.SEARCHING_ENTROPY = False
        if not sett.DONE and sett.NEW_DB:
            _remove_files()
        die()


@handle_keyboardinterrupt
def _secure():
    """Handle boltlight and implementation secrets."""
    getLogger(sett.PKG_NAME + '.errors').setLevel(CRITICAL)
    boltlight_password = environ.get('boltlight_password')
    try:
        if boltlight_password:
            # pylint: disable=consider-using-with
            sys.stdout = open(devnull, 'w', encoding='utf-8')
            # pylint: enable=consider-using-with
        init_common("Start boltlight's secure procedure", write_perms=True)
        if not _db_exists():
            sett.NEW_DB = True
        elif not _keep_db(not boltlight_password):
            sett.NEW_DB = True
        init_db(new_db=sett.NEW_DB)
        with session_scope(FakeContext()) as session:
            if boltlight_password:
                db_config_non_interactive(session, sett.NEW_DB,
                                          boltlight_password)
            else:
                db_config_interactive(session, sett.NEW_DB)
        sett.DONE = True
        print('All done!')
    finally:
        if boltlight_password:
            sys.stdout.close()



# Boltlight - a LN node wrapper
#
# Copyright (C) 2021-2022 boltlight contributors
# Copyright (C) 2018 inbitcoin s.r.l.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Miscellaneous utils module."""

import sys
from argparse import ArgumentParser
from configparser import ConfigParser
from contextlib import contextmanager
from distutils.util import strtobool
from functools import wraps
from importlib import import_module
from importlib.resources import files
from logging import CRITICAL, NOTSET, disable, getLogger
from logging.config import dictConfig
from os import R_OK, W_OK, access, mkdir, path
from pathlib import Path
from shutil import copyfile
from threading import current_thread

from .. import __version__
from .. import settings as sett
from ..migrate import migrate
from .exceptions import InterruptException

LOGGER = getLogger(__name__)


def handle_keyboardinterrupt(func):
    """Handle KeyboardInterrupt by raising an InterruptException."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except KeyboardInterrupt:
            print('\nKeyboard interrupt detected.')
            raise InterruptException from None

    return wrapper


def handle_thread(func):
    """Add and remove async threads from global list."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        sett.THREADS.append(current_thread())
        try:
            res = func(*args, **kwargs)
            sett.THREADS.remove(current_thread())
            return res
        except Exception as exc:
            sett.THREADS.remove(current_thread())
            raise exc

    return wrapper


@contextmanager
def disable_logger():
    """Disable logging.

    Warning: do not nest calls to this method.
    """
    disable(CRITICAL)
    try:
        yield
    finally:
        disable(NOTSET)


def copy_config_sample(interactive):
    """Copy (or ask to copy) config.sample and exit."""
    if interactive:
        copy = str2bool(input(
            'Missing configuration file, do you want a copy of '
            'config.sample in the specified location '
            f'({sett.CONFIG})? [Y/n] '),
                        force_true=True)
        if not copy:
            die("You'll need to manually create a configuration file")
    else:
        LOGGER.error(
            "Missing config file, copying sample to '%s', "
            'read doc/configuring.md for details', sett.CONFIG)
    sample = files(sett.PKG_NAME).joinpath('share/config.sample')
    try:
        copyfile(sample, sett.CONFIG)
    except OSError as err:
        die('Error copying sample file: ' + str(err))
    die('Please configure boltlight')


def die(message=None, exit_code=1):
    """Print message to stderr and exit with the requested error code."""
    if message:
        sys.stderr.write(message + '\n')
    sys.exit(exit_code)


def get_config_parser(interactive=False):
    """Read config file, set default values and return a config parser.

    When config is missing, copy config.sample in its expected location and
    terminate.
    """
    sett.CONFIG = path.join(sett.DATA, sett.CONFIG_NAME)
    if not path.exists(sett.CONFIG):
        copy_config_sample(interactive)
    config = ConfigParser()
    config.read(sett.CONFIG)
    l_values = [
        'INSECURE_CONNECTION', 'PORT', 'SERVER_KEY', 'SERVER_CRT', 'LOGS_DIR',
        'LOGS_LEVEL', 'DB_DIR', 'MACAROONS_DIR', 'DISABLE_MACAROONS'
    ]
    set_defaults(config, l_values)
    return config


def get_path(ipath, base_path=None):
    """Get absolute posix path.

    By default relative paths are calculated from boltlightdir.
    """
    ipath = Path(ipath).expanduser()
    if ipath.is_absolute():
        return ipath.as_posix()
    if not base_path:
        base_path = sett.DATA
    return Path(base_path, ipath).as_posix()


def handle_importerror(err):
    """Handle an ImportError."""
    LOGGER.debug('Import error: %s', str(err))
    LOGGER.error("Implementation '%s' is not supported", sett.IMPLEMENTATION)
    die()


def handle_sigterm(_signo, _stack_frame):
    """Handle a SIGTERM by raising an InterruptException."""
    raise InterruptException


def init_common(help_msg, core=True, write_perms=False, runtime=False):
    """Initialize common entrypoints calls."""
    _update_logger()
    _parse_args(help_msg, write_perms)  # updates sett.DATA
    if write_perms:
        if not access(sett.DATA, W_OK):
            raise RuntimeError('Permission denied on ' + sett.DATA)
    _init_tree()
    config = get_config_parser()  # updates sett.CONFIG
    _update_logger(config)
    _get_start_options(config, runtime)
    if core:
        migrate()
        # reupdating logger as migrate overrides configuration
        _update_logger(config)


def _update_logger(config=None):
    """Activate console logs.

    When configuration is available, activate file logs and set configured log
    level.
    """
    if config:
        sec = 'boltlight'
        logs_level = config.get(sec, 'LOGS_LEVEL').upper()
        sett.LOGGING['handlers']['console']['level'] = logs_level
        sett.LOGGING['loggers']['']['handlers'].append('file')
        sett.LOGGING['handlers'].update(sett.LOGGING_FILE)
        sett.LOGS_DIR = get_path(config.get(sec, 'LOGS_DIR'))
        log_path = path.join(sett.LOGS_DIR, sett.LOGS_BOLTLIGHT)
        sett.LOGGING['handlers']['file']['filename'] = log_path
    try:
        dictConfig(sett.LOGGING)
    except (AttributeError, ImportError, TypeError, ValueError) as err:
        raise RuntimeError('Logging configuration error: ' +
                           str(err)) from None
    getLogger('urllib3').propagate = False


def _parse_args(help_msg, write_perms):
    """Parse command line arguments."""
    parser = ArgumentParser(description=help_msg)
    acc_mode = R_OK
    if write_perms:
        acc_mode = W_OK
    parser.add_argument('--boltlightdir',
                        metavar='PATH',
                        help="Path containing config file and other data")
    args = vars(parser.parse_args())
    if 'boltlightdir' in args and args['boltlightdir'] is not None:
        boltlightdir = args['boltlightdir']
        if not boltlightdir:
            raise RuntimeError('Invalid boltlightdir: empty path')
        if not path.isdir(boltlightdir):
            raise RuntimeError('Invalid boltlightdir: path is not a directory')
        if not access(boltlightdir, acc_mode):
            raise RuntimeError('Invalid boltlightdir: permission denied')
        sett.DATA = boltlightdir


def _init_tree():
    """Create data directory tree if missing."""
    _try_mkdir(sett.DATA)
    _try_mkdir(path.join(sett.DATA, 'certs'))
    _try_mkdir(path.join(sett.DATA, 'db'))
    _try_mkdir(path.join(sett.DATA, 'logs'))
    _try_mkdir(path.join(sett.DATA, 'macaroons'))


def _try_mkdir(dir_path):
    """Create a directory if it doesn't exist."""
    if not path.exists(dir_path):
        LOGGER.info('Creating dir %s', dir_path)
        mkdir(dir_path)


def _get_start_options(config, runtime):
    """Set boltlight and implementation start options."""
    sec = 'boltlight'
    sett.IMPLEMENTATION = config.get(sec, 'IMPLEMENTATION').lower()
    sett.INSECURE_CONNECTION = str2bool(config.get(sec, 'INSECURE_CONNECTION'))
    sett.DISABLE_MACAROONS = str2bool(config.get(sec, 'DISABLE_MACAROONS'))
    sett.PORT = config.get(sec, 'PORT')
    sett.LISTEN_ADDR = f'{sett.HOST}:{sett.PORT}'
    if sett.INSECURE_CONNECTION:
        sett.DISABLE_MACAROONS = True
    sett.SERVER_KEY = get_path(config.get(sec, 'SERVER_KEY'))
    sett.SERVER_CRT = get_path(config.get(sec, 'SERVER_CRT'))
    if sett.DISABLE_MACAROONS:
        LOGGER.warning('Disabling macaroons is not safe, '
                       'do not disable them in production')
    sett.MACAROONS_DIR = get_path(config.get(sec, 'MACAROONS_DIR'))
    sett.DB_DIR = get_path(config.get(sec, 'DB_DIR'))
    sett.DB_PATH = path.join(sett.DB_DIR, sett.DB_NAME)
    # Checks if implementation is supported, could throw an ImportError
    module = import_module(f'...light_{sett.IMPLEMENTATION}', __name__)
    if runtime:
        getattr(module, 'get_settings')(config, sett.IMPLEMENTATION)


def set_defaults(config, values):
    """Set configuration defaults."""
    defaults = {}
    for var in values:
        defaults[var] = getattr(sett, var)
    config.read_dict({'DEFAULT': defaults})


def str2bool(string, force_true=False):
    """Cast a string to boolean, forcing to a default value."""
    try:
        return strtobool(str(string).lower())
    except ValueError:
        return force_true



# Boltlight - a LN node wrapper
#
# Copyright (C) 2021-2022 boltlight contributors
# Copyright (C) 2018 inbitcoin s.r.l.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Settings module for boltlight.

WARNING: only Python's core module imports here, to avoid breaking build system
         and import loops.
"""

from os import path
from pathlib import Path

# Empty variables are set at runtime
# Some variables contain default values, could be overwritten

IMPLEMENTATION = ''

DATA = path.join(str(Path.home()), '.boltlight')
CONFIG_NAME = 'config'
CONFIG = ''

PKG_NAME = 'boltlight'
PIP_NAME = 'boltlight'

HOST = '0.0.0.0'
PORT = '1708'
LISTEN_ADDR = ''
INSECURE_CONNECTION = 0
SERVER_KEY = './certs/server.key'
SERVER_CRT = './certs/server.crt'
IMPLEMENTATION_SECRETS = False
IMPL_SEC_TYPE = ''

# Macaroons settings
RUNTIME_BAKER = None
DISABLE_MACAROONS = 0
MACAROONS_DIR = 'macaroons'
MAC_ADMIN = 'admin.macaroon'
MAC_READONLY = 'readonly.macaroon'
MAC_INVOICES = 'invoices.macaroon'

# Security settings
MAC_ROOT_KEY = None
SALT_LEN = 32
ACCESS_TOKEN = b'boltlight'
PASSWORD_LEN = 12
SCRYPT_PARAMS = {
    'cost_factor': 2**15,
    'block_size_factor': 8,
    'parallelization_factor': 1,
    'key_len': 32
}

# boltlight-secure settings
COLLECTING_INPUT = True
DONE = False
NEW_DB = False
SEARCHING_ENTROPY = True
FIRST_WORK_TIME = None
IDLE_COUNTER = 1
IDLE_MESSAGES = {
    1: {
        'msg': 'please keep generating entropy',
        'delay': 5
    },
    2: {
        'msg': 'more entropy, please',
        'delay': 15
    },
    3: {
        'msg': '...good things come to those who wait...',
        'delay': 30
    }
}

# DB settings
DB_DIR = 'db'
DB_NAME = 'boltlight.db'
DB_PATH = ''
ALEMBIC_CFG = path.join(path.dirname(__file__), 'migrations/alembic.ini')

# Server settings
GRPC_WORKERS = 10
GRPC_GRACE_TIME = 40
UNLOCKER_STOP = False
RUNTIME_STOP = False
RUNTIME_SERVER = None
THREADS = []

# blink settings
RPCSERVER = 'localhost:1708'
TLSCERT = './certs/server.crt'
MACAROON = path.join(MACAROONS_DIR, MAC_ADMIN)
INSECURE = 0
NO_MACAROON = 0
CLI_RPCSERVER = None
CLI_TLSCERT = None
CLI_MACAROON = None
CLI_INSECURE = None
CLI_NO_MACAROON = None
CLI_TIMEOUT = 10
CLI_BASE_GRPC_CODE = 64

ENFORCE = True

TEST_HASH = '43497fd7f826957108f4a30fd9cec3aeba79972084e90ead01ea330900000000'
MAIN_HASH = '6fe28c0ab6f1b372c1a6a246ae63f74f931e8365e15a089c68d6190000000000'

# RPC-based implementations settings
RPC_URL = ''
RPC_TRIES = 5
RPC_SLEEP = .5
RPC_CONN_TIMEOUT = 3.1

# c-lightning specific settings
CL_RPC = 'lightning-rpc'

# eclair specific settings
ECL_HOST = 'localhost'
ECL_PORT = 8080
ECL_PASS = ''
ECL_CONFIRMATIONS = 6

# electrum specific settings
ELE_HOST = 'localhost'
ELE_PORT = 7777
ELE_USER = 'user'
ELE_RELEASED_ADDRESSES = []

# lnd specific settings
LND_HOST = 'localhost'
LND_PORT = 10009
LND_CERT = 'tls.cert'
LND_ADDR = ''
LND_CREDS_SSL = ''
LND_CREDS_FULL = ''
LND_MAC = ''

# Common settings
IMPL_MIN_TIMEOUT = 2
IMPL_MAX_TIMEOUT = 180
RESPONSE_RESERVED_TIME = 0.3
THREAD_TIMEOUT = 3
CLOSE_TIMEOUT_NODE = 15
MAX_INVOICES = 200
INVOICES_TIMES = 3
EXPIRY_TIME = 420
DUST_LIMIT_SAT = 546

# Logging settings
LOGS_DIR = 'logs'
LOGS_BOLTLIGHT = 'boltlight.log'
LOGS_MIGRATIONS = 'migrations.log'
LOG_TIMEFMT = '%Y-%m-%d %H:%M:%S %z'
LOG_TIMEFMT_SIMPLE = '%d %b %H:%M:%S'
LOGS_LEVEL = 'INFO'
LOG_LEVEL_FILE = 'DEBUG'
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose': {
            'format':
            "[%(asctime)s] %(levelname).3s [%(name)s:%(lineno)s] %(message)s",
            'datefmt': LOG_TIMEFMT
        },
        'simple': {
            'format': '%(asctime)s %(levelname).3s: %(message)s',
            'datefmt': LOG_TIMEFMT_SIMPLE
        },
    },
    'handlers': {
        'console': {
            'level': LOGS_LEVEL,
            'class': 'logging.StreamHandler',
            'formatter': 'simple',
            'stream': 'ext://sys.stdout'
        },
    },
    'loggers': {
        '': {
            'handlers': ['console'],
            'level': 'DEBUG'
        },
    }
}
LOGGING_FILE = {
    'file': {
        'level': LOG_LEVEL_FILE,
        'class': 'logging.handlers.RotatingFileHandler',
        'filename': path.join(DATA, LOGS_DIR, LOGS_BOLTLIGHT),
        'maxBytes': 1048576,
        'backupCount': 7,
        'formatter': 'verbose'
    }
}

# Macaroons settings
BOLTLIGHT_PERMS = {
    '/boltlight.Boltlight/GetInfo': {
        'entity': 'info',
        'action': 'read'
    },
    '/boltlight.Boltlight/Lock': {
        'entity': 'lock',
        'action': 'write'
    },
}
LIGHTNING_PERMS = {
    '/boltlight.Lightning/BalanceOffChain': {
        'entity': 'balance',
        'action': 'read'
    },
    '/boltlight.Lightning/CheckInvoice': {
        'entity': 'invoice',
        'action': 'read'
    },
    '/boltlight.Lightning/CloseChannel': {
        'entity': 'channel',
        'action': 'write'
    },
    '/boltlight.Lightning/CreateInvoice': {
        'entity': 'invoice',
        'action': 'write'
    },
    '/boltlight.Lightning/DecodeInvoice': {
        'entity': 'invoice',
        'action': 'read'
    },
    '/boltlight.Lightning/GetNodeInfo': {
        'entity': 'info',
        'action': 'read'
    },
    '/boltlight.Lightning/ListChannels': {
        'entity': 'channel',
        'action': 'read'
    },
    '/boltlight.Lightning/ListInvoices': {
        'entity': 'invoice',
        'action': 'read'
    },
    '/boltlight.Lightning/ListPayments': {
        'entity': 'payment',
        'action': 'read'
    },
    '/boltlight.Lightning/ListPeers': {
        'entity': 'peer',
        'action': 'read'
    },
    '/boltlight.Lightning/ListTransactions': {
        'entity': 'transaction',
        'action': 'read'
    },
    '/boltlight.Lightning/NewAddress': {
        'entity': 'address',
        'action': 'write'
    },
    '/boltlight.Lightning/OpenChannel': {
        'entity': 'channel',
        'action': 'write'
    },
    '/boltlight.Lightning/PayInvoice': {
        'entity': 'payment',
        'action': 'write'
    },
    '/boltlight.Lightning/PayOnChain': {
        'entity': 'transaction',
        'action': 'write'
    },
    '/boltlight.Lightning/UnlockNode': {
        'entity': 'unlock',
        'action': 'write'
    },
    '/boltlight.Lightning/BalanceOnChain': {
        'entity': 'balance',
        'action': 'read'
    },
}
ALL_PERMS = {**BOLTLIGHT_PERMS, **LIGHTNING_PERMS}
READ_PERMS = [
    {
        'entity': 'balance',
        'action': 'read'
    },
    {
        'entity': 'channel',
        'action': 'read'
    },
    {
        'entity': 'info',
        'action': 'read'
    },
    {
        'entity': 'invoice',
        'action': 'read'
    },
    {
        'entity': 'payment',
        'action': 'read'
    },
    {
        'entity': 'peer',
        'action': 'read'
    },
    {
        'entity': 'transaction',
        'action': 'read'
    },
]
INVOICE_PERMS = [
    {
        'entity': 'channel',
        'action': 'read'
    },
    {
        'entity': 'info',
        'action': 'read'
    },
    {
        'entity': 'invoice',
        'action': 'read'
    },
    {
        'entity': 'invoice',
        'action': 'write'
    },
    {
        'entity': 'peer',
        'action': 'read'
    },
]



# Boltlight - a LN node wrapper
#
# Copyright (C) 2021-2022 boltlight contributors
# Copyright (C) 2018 inbitcoin s.r.l.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Macaroons management (creation and validation) class."""

from codecs import decode
from logging import getLogger

from macaroonbakery.bakery import (
    LATEST_VERSION, AuthInitError, Bakery, DischargeRequiredError,
    MemoryKeyStore, MemoryOpsStore, Op, PermissionDenied, canonical_ops)
from macaroonbakery.checkers import AuthContext, context_with_operations
from pymacaroons import Macaroon
from pymacaroons.exceptions import MacaroonException

from . import settings as sett
from .utils.boltlight import RuntimeTerminate

LOGGER = getLogger(__name__)

MAC_VERSION = LATEST_VERSION

ALL_OPS = canonical_ops(
    [Op(op['entity'], op['action']) for op in sett.ALL_PERMS.values()])

INVOICE_OPS = canonical_ops(
    [Op(op['entity'], op['action']) for op in sett.INVOICE_PERMS])

READONLY_OPS = canonical_ops(
    [Op(op['entity'], op['action']) for op in sett.READ_PERMS])

MACAROONS = {
    sett.MAC_ADMIN: ALL_OPS,
    sett.MAC_INVOICES: INVOICE_OPS,
    sett.MAC_READONLY: READONLY_OPS,
}


def check_macaroons(metadata, method):
    """Check if metadata contains valid macaroons.

    Return None if authorization is granted, error handler method otherwise.
    """
    num_mac = 0
    for data in metadata:
        if data.key == 'macaroon':
            num_mac = num_mac + 1
            try:
                serialized_macaroon = decode(data.value, 'hex')
                macaroon = Macaroon.deserialize(serialized_macaroon)
            except (MacaroonException, ValueError):
                LOGGER.error('- Cannot deserialize macaroon')
                return RuntimeTerminate.macaroon_error
    if num_mac == 0:
        LOGGER.error('- No macaroons were provided')
        return RuntimeTerminate.missing_macaroon
    if num_mac > 1:
        LOGGER.error('- Wrong number of macaroons, 1 required, %s received',
                     num_mac)
        return RuntimeTerminate.macaroon_error
    if _validate_macaroon(macaroon, sett.ALL_PERMS[method]):
        return None
    LOGGER.error('- Macaroon does not gran permission for this operation')
    return RuntimeTerminate.macaroon_error


def _validate_macaroon(macaroon, required_perm):
    """Check if a given macaroon is authorized to run required operation."""
    baker = sett.RUNTIME_BAKER
    auth_checker = baker.checker.auth([[macaroon]])
    ctx_op = context_with_operations(AuthContext(), ALL_OPS)
    required_op = Op(required_perm['entity'], required_perm['action'])
    try:
        auth_info = auth_checker.allow(ctx_op, [required_op])
        if auth_info:
            return True
    except (AuthInitError, DischargeRequiredError, PermissionDenied):
        LOGGER.error('- Authorization error')
    return False


def get_baker(root_key, put_ops=False):
    """Get a baker, optionally registering operations in MemoryOpsStore."""
    baker = Bakery(location='boltlight',
                   ops_store=MemoryOpsStore(),
                   root_key_store=MemoryKeyStore(key=root_key))
    if put_ops:
        for permitted_ops in MACAROONS.values():
            entity = baker.oven.ops_entity(permitted_ops)
            baker.oven.ops_store.put_ops(entity, None, permitted_ops)
    return baker



# Boltlight - a LN node wrapper
#
# Copyright (C) 2021-2022 boltlight contributors
# Copyright (C) 2018 inbitcoin s.r.l.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For a full list of contributors, please see the AUTHORS.md file.
"""Module that invokes the migration engine, based on generated sample."""

from logging import getLogger
from logging.config import fileConfig

from alembic import context
from sqlalchemy import engine_from_config, pool

CFG = context.config

SEC = 'boltlight_log'
env = {}
env['boltlightlogfilename'] = 'boltlight.log'
if CFG.get_section_option(SEC, 'boltlight'):
    env['boltlightlogfilename'] = CFG.get_section_option(SEC, 'boltlight')
env['migrationslogfilename'] = 'migrations.log'
if CFG.get_section_option(SEC, 'migrations'):
    env['migrationslogfilename'] = CFG.get_section_option(SEC, 'migrations')

fileConfig(CFG.config_file_name, disable_existing_loggers=False, defaults=env)

LOGGER = getLogger(__name__)

LOGGER.info('%s DB migration: begin %s', '*' * 25, '*' * 26)

# add your model's MetaData object here for 'autogenerate' support
# from myapp import mymodel
# target_metadata = mymodel.Base.metadata
TARGET_METADATA = None


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL and not an Engine, though an
    Engine is acceptable here as well. By skipping the Engine creation we don't
    even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the script output.
    """
    print('Offline mode is not supported')
    import sys
    sys.exit(1)


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine and associate a connection
    with the context.
    """
    try:
        connectable = CFG.attributes.get('connection', None)
        if connectable is None:
            # only create Engine if we don't have a Connection
            # from the outside
            connectable = engine_from_config(
                CFG.get_section(CFG.config_ini_section),
                prefix="sqlalchemy.",
                poolclass=pool.NullPool,
            )
        with connectable.connect() as connection:
            context.configure(connection=connection,
                              TARGET_METADATA=TARGET_METADATA)
            with context.begin_transaction():
                context.run_migrations()
    except Exception:
        LOGGER.error('Error during DB migration')
        import sys
        sys.exit(1)


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()

LOGGER.info('%s DB migration: end %s', '*' * 25, '*' * 28)



# Boltlight - a LN node wrapper
#
# Copyright (C) 2021-2022 boltlight contributors
# Copyright (C) 2018 inbitcoin s.r.l.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For a full list of contributors, please see the AUTHORS.md file.
"""Implementation of boltlight.proto defined methods for lnd."""

from binascii import hexlify, unhexlify
from codecs import encode
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import TimeoutError as TimeoutFutError
from contextlib import ExitStack, contextmanager, suppress
from datetime import datetime
from functools import wraps
from logging import getLogger
from os import path

from grpc import (
    FutureTimeoutError, RpcError, channel_ready_future,
    composite_channel_credentials, metadata_call_credentials, secure_channel,
    ssl_channel_credentials)
from lnd_proto import rpc_pb2 as ln
from lnd_proto import rpc_pb2_grpc as lnrpc
from lnd_proto import walletunlocker_pb2 as wu_ln
from lnd_proto import walletunlocker_pb2_grpc as wu_lnrpc

from . import boltlight_pb2 as pb
from . import settings as sett
from .errors import Err
from .utils.bitcoin import Enforcer as Enf
from .utils.bitcoin import convert, get_channel_balances, has_amount_encoded
from .utils.db import session_scope
from .utils.misc import get_path, handle_thread, set_defaults
from .utils.network import (
    FakeContext, check_req_params, get_node_timeout, get_thread_timeout)
from .utils.security import get_secret, unlock_node_with_password

LOGGER = getLogger(__name__)

ERRORS = {
    'already connected to peer': {
        'fun': 'connect_failed'
    },
    'amount must be specified when paying a zero amount invoice': {
        'fun': 'amount_required'
    },
    'chain backend is still syncing': {
        'fun': 'node_error'
    },
    'channels cannot be created before the wallet is fully synced': {
        'fun': 'openchannel_failed'
    },
    'checksum failed': {
        'fun': 'invalid',
        'params': 'payment_request'
    },
    'checksum mismatch': {
        'fun': 'invalid',
        'params': 'address'
    },
    'Deadline Exceeded': {
        'fun': 'node_error'
    },
    'decoded address is of unknown format': {
        'fun': 'invalid',
        'params': 'address'
    },
    'edge not found': {
        'fun': 'invalid',
        'params': 'channel_id'
    },
    'encoding/hex': {
        'fun': 'invalid',
        'params': 'payment_hash'
    },
    'expected 1 macaroon, got': {
        'fun': 'node_error'
    },
    'greater than max expiry of': {
        'fun': 'invalid',
        'params': 'expiry'
    },
    'i/o timeout': {
        'fun': 'node_error'
    },
    'insufficient_balance': {
        'fun': 'insufficient_funds'
    },
    'invalid bech32 string length': {
        'fun': 'invalid',
        'params': 'payment_request'
    },
    'invalid index of': {
        'fun': 'invalid',
        'params': 'payment_request'
    },
    'invoice expired': {
        'fun': 'invoice_expired'
    },
    'invalid funding_satoshis': {
        'fun': 'openchannel_failed'
    },
    'invalid passphrase for master public key': {
        'fun': 'wrong_node_password'
    },
    'invoice is already paid': {
        'fun': 'invalid',
        'params': 'payment_request'
    },
    'is not online': {
        'fun': 'connect_failed'
    },
    'Name resolution failure': {
        'fun': 'node_error'
    },
    'no_route': {
        'fun': 'payinvoice_failed'
    },
    'Number of pending channels exceed maximum': {
        'fun': 'openchannel_failed'
    },
    'payment hash must': {
        'fun': 'invalid',
        'params': 'payment_hash'
    },
    'received funding error from': {
        'fun': 'openchannel_failed'
    },
    'signature mismatch': {
        'fun': 'node_error'
    },
    'Socket closed': {
        'fun': 'node_error'
    },
    'string not all lowercase or all uppercase': {
        'fun': 'invalid',
        'params': 'payment_request'
    },
    'unable to find a path to destination': {
        'fun': 'route_not_found'
    },
    'unable to find arbitrator': {
        'fun': 'closechannel_failed'
    },
    'unable to get best block info': {
        'fun': 'node_error'
    },
    'unable to gracefully close channel while peer is offline': {
        'fun': 'closechannel_failed'
    },
    'unable to locate invoice': {
        'fun': 'invoice_not_found'
    },
    'unable to route payment to destination: FeeInsufficient': {
        'fun': 'insufficient_fee'
    },
    'unable to route payment to destination: TemporaryChannelFailure': {
        'fun': 'payinvoice_failed'
    },
    'unknown service lnrpc.Lightning': {
        'fun': 'node_locked'
    }
}

LND_LN_TX = {'min_value': 1, 'max_value': 2**32 / 1000, 'unit': Enf.SATS}
LND_PUSH = {'min_value': 0, 'max_value': 2**24, 'unit': Enf.SATS}


def get_node_version():
    """Get node's version."""
    with suppress(RuntimeError):
        with _connect(FakeContext()) as stub:
            with suppress(RpcError):
                lnd_res = stub.GetInfo(ln.GetInfoRequest(),
                                       timeout=sett.IMPL_MIN_TIMEOUT)
                return lnd_res.version
    return ''


def get_settings(config, sec):
    """Get lnd settings."""
    sett.IMPL_SEC_TYPE = 'macaroon'
    lnd_values = ['LND_HOST', 'LND_PORT', 'LND_CERT']
    set_defaults(config, lnd_values)
    lnd_host = config.get(sec, 'LND_HOST')
    lnd_port = config.get(sec, 'LND_PORT')
    sett.LND_ADDR = f'{lnd_host}:{lnd_port}'
    lnd_tls_cert_dir = get_path(config.get(sec, 'LND_CERT_DIR'))
    lnd_tls_cert = config.get(sec, 'LND_CERT')
    lnd_tls_cert_path = path.join(lnd_tls_cert_dir, lnd_tls_cert)
    with open(lnd_tls_cert_path, 'rb') as file:
        cert = file.read()
    # Build ssl credentials using the cert
    sett.LND_CREDS_FULL = sett.LND_CREDS_SSL = ssl_channel_credentials(cert)


def unlock_node(ctx, password, session=None):
    """Unlock node with password saved in boltlight's DB."""
    with ExitStack() if session else session_scope(ctx) as ses:
        if session:
            ses = session
        lnd_pass = get_secret(ctx, ses, password, 'lnd', 'password')
        if not lnd_pass:
            Err().node_error(
                ctx, 'No password stored, add one by '
                'running boltlight-secure')
        lnd_req = wu_ln.UnlockWalletRequest(wallet_password=lnd_pass)
        try:
            with _connect(ctx,
                          stub_class=wu_lnrpc.WalletUnlockerStub,
                          force_no_macaroon=True) as stub:
                stub.UnlockWallet(lnd_req, timeout=get_node_timeout(ctx))
        except RpcError as err:
            if 'wallet already unlocked' in str(err):
                LOGGER.info('Node is already unlocked')
            else:
                _handle_error(ctx, str(err))


def update_settings(macaroon):
    """Update lnd specific settings."""
    if macaroon:
        LOGGER.info('Connecting to lnd in secure mode (tls + macaroon)')
        sett.LND_MAC = macaroon
        # Build meta data credentials
        auth_creds = metadata_call_credentials(_metadata_callback)
        # Combine the cert credentials and the macaroon auth credentials
        # Such that every call is properly encrypted and authenticated
        sett.LND_CREDS_FULL = composite_channel_credentials(
            sett.LND_CREDS_SSL, auth_creds)
    else:
        LOGGER.info('Connecting to lnd in insecure mode')


def _handle_rpc_errors(func):
    """Catch gRPC errors."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except RpcError as error:
            return _handle_error(args[1], error)

    return wrapper


@_handle_rpc_errors
def BalanceOffChain(_req, ctx):
    """Return the off-chain balance available across all channels."""
    channels = ListChannels(pb.ListChannelsRequest(), ctx).channels
    return get_channel_balances(channels)


@_handle_rpc_errors
def BalanceOnChain(_req, ctx):
    """Return the on-chain balance in satoshi of the running LN node."""
    res = pb.BalanceOnChainResponse()
    lnd_req = ln.WalletBalanceRequest()
    with _connect(ctx) as stub:
        lnd_res = stub.WalletBalance(lnd_req, timeout=get_node_timeout(ctx))
        res = pb.BalanceOnChainResponse(
            total_sat=lnd_res.total_balance,
            confirmed_sat=lnd_res.confirmed_balance)
    return res


@_handle_rpc_errors
def CheckInvoice(req, ctx):
    """Check if a LN invoice has been paid."""
    check_req_params(ctx, req, 'payment_hash')
    res = pb.CheckInvoiceResponse()
    lnd_req = ln.PaymentHash(r_hash=unhexlify(req.payment_hash))
    with _connect(ctx) as stub:
        lnd_res = stub.LookupInvoice(lnd_req, timeout=get_node_timeout(ctx))
        res.state = _get_invoice_state(lnd_res)
    return res


@_handle_rpc_errors
def CloseChannel(req, ctx):
    """Try to close a LN chanel."""
    check_req_params(ctx, req, 'channel_id')
    channel_id = 0
    try:
        channel_id = int(req.channel_id)
    except ValueError:
        Err().invalid(ctx, 'channel_id')
    lnd_req = ln.ChanInfoRequest(chan_id=channel_id)
    with _connect(ctx) as stub:
        lnd_res = stub.GetChanInfo(lnd_req, timeout=get_node_timeout(ctx))
        txid, vout = lnd_res.chan_point.split(':')
        chan_point = ln.ChannelPoint(funding_txid_str=txid,
                                     output_index=int(vout))
        lnd_req = ln.CloseChannelRequest(channel_point=chan_point,
                                         force=req.force)
        with ThreadPoolExecutor(max_workers=1) as executor:
            close_time = get_node_timeout(ctx,
                                          min_time=sett.CLOSE_TIMEOUT_NODE)
            future = executor.submit(_close_channel, lnd_req, close_time)
            try:
                lnd_res = future.result(timeout=get_thread_timeout(ctx))
                if lnd_res:
                    return pb.CloseChannelResponse(closing_txid=lnd_res)
            except TimeoutFutError:
                executor.shutdown(wait=False)
            except RpcError as err:
                _handle_error(ctx, err)
            except RuntimeError as err:
                _handle_error(ctx, str(err))
    return pb.CloseChannelResponse()


@_handle_rpc_errors
def CreateInvoice(req, ctx):
    """Create a LN invoice (bolt 11)."""
    expiry = sett.EXPIRY_TIME
    if req.expiry:
        expiry = req.expiry
    lnd_req = ln.Invoice(memo=req.description,
                         expiry=expiry,
                         fallback_addr=req.fallback_addr)
    if req.min_final_cltv_expiry:
        Enf.check_value(ctx,
                        req.min_final_cltv_expiry,
                        enforce=Enf.MIN_FINAL_CLTV_EXPIRY)
        lnd_req.cltv_expiry = req.min_final_cltv_expiry
    if req.amount_msat:
        lnd_req.value_msat = req.amount_msat
    res = pb.CreateInvoiceResponse()
    with _connect(ctx) as stub:
        lnd_res = stub.AddInvoice(lnd_req, timeout=get_node_timeout(ctx))
        payment_hash_str = ''
        if lnd_res.r_hash:
            payment_hash_str = hexlify(lnd_res.r_hash).decode()
        res.payment_hash = payment_hash_str
        res.payment_request = lnd_res.payment_request
        if payment_hash_str:
            lnd_req = ln.PaymentHash(r_hash=lnd_res.r_hash)
            lnd_res = stub.LookupInvoice(lnd_req,
                                         timeout=get_node_timeout(ctx))
            res.expires_at = lnd_res.creation_date + lnd_res.expiry
    return res


@_handle_rpc_errors
def DecodeInvoice(req, ctx):
    """Return information of an invoice from its payment request (bolt 11)."""
    check_req_params(ctx, req, 'payment_request')
    if req.description:
        Err().unimplemented_parameter(ctx, 'description')
    lnd_req = ln.PayReqString(pay_req=req.payment_request)
    res = pb.DecodeInvoiceResponse()
    with _connect(ctx) as stub:
        lnd_res = stub.DecodePayReq(lnd_req, timeout=get_node_timeout(ctx))
        res.amount_msat = lnd_res.num_satoshis * 1000
        res.timestamp = lnd_res.timestamp
        res.payment_hash = lnd_res.payment_hash
        res.description = lnd_res.description
        res.destination_pubkey = lnd_res.destination
        res.description_hash = lnd_res.description_hash
        res.expiry = lnd_res.expiry
        res.min_final_cltv_expiry = lnd_res.cltv_expiry
        res.fallback_addr = lnd_res.fallback_addr
        for lnd_route in lnd_res.route_hints:
            _add_route_hint(res, lnd_route)
    return res


@_handle_rpc_errors
def GetNodeInfo(_req, ctx):
    """Return info about the running LN node."""
    res = pb.GetNodeInfoResponse()
    lnd_req = ln.GetInfoRequest()
    with _connect(ctx) as stub:
        lnd_res = stub.GetInfo(lnd_req, timeout=get_node_timeout(ctx))
        res = pb.GetNodeInfoResponse(identity_pubkey=lnd_res.identity_pubkey,
                                     alias=lnd_res.alias,
                                     color=lnd_res.color,
                                     block_height=lnd_res.block_height)
        if lnd_res.chains[0].network == 'mainnet':
            res.network = pb.Network.MAINNET
        elif lnd_res.chains[0].network == 'testnet':
            res.network = pb.Network.TESTNET
        elif lnd_res.chains[0].network == 'regtest':
            res.network = pb.Network.REGTEST
        if lnd_res.uris:
            res.node_uri = lnd_res.uris[0]
    return res


@_handle_rpc_errors
def ListChannels(req, ctx):
    """Return a list of channels of the running LN node."""
    res = pb.ListChannelsResponse()
    lnd_req = ln.ListChannelsRequest()
    with _connect(ctx) as stub:
        lnd_res = stub.ListChannels(lnd_req, timeout=get_node_timeout(ctx))
        for lnd_chan in lnd_res.channels:
            _add_channel(res,
                         lnd_chan,
                         pb.Channel.OPEN,
                         active_only=req.active_only,
                         open_chan=True)
        if not req.active_only:
            lnd_req = ln.PendingChannelsRequest()
            lnd_res = stub.PendingChannels(lnd_req,
                                           timeout=get_node_timeout(ctx))
            for lnd_chan in lnd_res.pending_open_channels:
                _add_channel(res, lnd_chan, pb.Channel.PENDING_OPEN)
            for lnd_chan in lnd_res.pending_closing_channels:
                _add_channel(res, lnd_chan, pb.Channel.PENDING_MUTUAL_CLOSE)
            for lnd_chan in lnd_res.pending_force_closing_channels:
                _add_channel(res, lnd_chan, pb.Channel.PENDING_FORCE_CLOSE)
            for lnd_chan in lnd_res.waiting_close_channels:
                _add_channel(res, lnd_chan, pb.Channel.UNKNOWN)
    return res


@_handle_rpc_errors
def ListInvoices(req, ctx):
    """Return a list of lightning invoices created by the running LN node."""
    if not req.max_items:
        req.max_items = sett.MAX_INVOICES
    res = pb.ListInvoicesResponse()
    lnd_req = ln.ListInvoiceRequest(reversed=req.search_order,
                                    num_max_invoices=req.max_items *
                                    sett.INVOICES_TIMES)
    stop = False
    with _connect(ctx) as stub:
        while True:
            lnd_res = stub.ListInvoices(lnd_req, timeout=get_node_timeout(ctx))
            if not lnd_res.invoices:
                break
            if req.search_order:
                lnd_res.CopyFrom(
                    ln.ListInvoiceResponse(
                        invoices=reversed(lnd_res.invoices),
                        first_index_offset=lnd_res.first_index_offset,
                        last_index_offset=lnd_res.last_index_offset))
            stop = _parse_invoices(res, lnd_res.invoices, req)
            if stop:
                break
            if req.search_order:
                lnd_req.index_offset = lnd_res.first_index_offset
            else:
                lnd_req.index_offset = lnd_res.last_index_offset
        if req.list_order != req.search_order:
            res.CopyFrom(
                pb.ListInvoicesResponse(invoices=reversed(res.invoices)))
    return res


@_handle_rpc_errors
def ListPayments(_req, ctx):
    """Return a list of lightning invoices paid by the running LN node."""
    res = pb.ListPaymentsResponse()
    lnd_req = ln.ListPaymentsRequest()
    with _connect(ctx) as stub:
        lnd_res = stub.ListPayments(lnd_req, timeout=get_node_timeout(ctx))
        for lnd_payment in lnd_res.payments:
            _add_payment(res, lnd_payment)
    return res


@_handle_rpc_errors
def ListPeers(_req, ctx):
    """Return a list of peers connected to the running LN node."""
    res = pb.ListPeersResponse()
    lnd_req = ln.ListPeersRequest()
    with _connect(ctx) as stub:
        lnd_res = stub.ListPeers(lnd_req, timeout=get_node_timeout(ctx))
        for lnd_peer in lnd_res.peers:
            peer = res.peers.add(pubkey=lnd_peer.pub_key,
                                 address=lnd_peer.address)
            lnd_req = ln.NodeInfoRequest(pub_key=lnd_peer.pub_key)
            with suppress(RpcError):
                lnd_res = stub.GetNodeInfo(lnd_req,
                                           timeout=get_node_timeout(ctx))
                peer.alias = lnd_res.node.alias
                peer.color = lnd_res.node.color
    return res


@_handle_rpc_errors
def ListTransactions(_req, ctx):
    """Return a list of on-chain transactions of the running LN node."""
    res = pb.ListTransactionsResponse()
    lnd_req = ln.GetTransactionsRequest()
    with _connect(ctx) as stub:
        lnd_res = stub.GetTransactions(lnd_req, timeout=get_node_timeout(ctx))
        for lnd_transaction in lnd_res.transactions:
            _add_transaction(res, lnd_transaction)
    return res


@_handle_rpc_errors
def NewAddress(req, ctx):
    """Create a new bitcoin address under control of the running LN node."""
    res = pb.NewAddressResponse()
    if req.addr_type == pb.Address.P2WPKH:
        # in lnd WITNESS_PUBKEY_HASH = 0;
        lnd_req = ln.NewAddressRequest(type=0)
    else:
        # in lnd NESTED_PUBKEY_HASH = 1;
        lnd_req = ln.NewAddressRequest(type=1)
    with _connect(ctx) as stub:
        lnd_res = stub.NewAddress(lnd_req, timeout=get_node_timeout(ctx))
        res = pb.NewAddressResponse(address=lnd_res.address)
    return res


@_handle_rpc_errors
def OpenChannel(req, ctx):
    """Try to connect and open a channel with a peer."""
    res = pb.OpenChannelResponse()
    check_req_params(ctx, req, 'node_uri', 'funding_sat')
    try:
        pubkey, host = req.node_uri.split('@')
    except ValueError:
        Err().invalid(ctx, 'node_uri')
    peer_address = ln.LightningAddress(pubkey=pubkey, host=host)
    lnd_req = ln.ConnectPeerRequest(addr=peer_address, perm=True)
    with _connect(ctx) as stub:
        try:
            lnd_res = stub.ConnectPeer(lnd_req, timeout=get_node_timeout(ctx))
        except RpcError as err:
            # pylint: disable=no-member
            if 'already connected to peer' not in err.details():
                # pylint: enable=no-member
                Err().connect_failed(ctx)
        Enf.check_value(ctx, req.funding_sat, Enf.FUNDING_SATOSHIS)
        lnd_req = ln.OpenChannelRequest(node_pubkey=unhexlify(pubkey),
                                        private=req.private,
                                        local_funding_amount=req.funding_sat)
        if req.push_msat:
            if req.push_msat >= 1000 * req.funding_sat:
                Err().value_too_high(ctx, req.push_msat)
            lnd_req.push_sat = convert(ctx,
                                       Enf.MSATS,
                                       Enf.SATS,
                                       req.push_msat,
                                       Enf.SATS,
                                       enforce=LND_PUSH)
        lnd_res = stub.OpenChannelSync(lnd_req, timeout=get_node_timeout(ctx))
        res.funding_txid = lnd_res.funding_txid_str
        if not lnd_res.funding_txid_str:
            txid = _txid_bytes_to_str(lnd_res.funding_txid_bytes)
            res.funding_txid = txid
    return res


@_handle_rpc_errors
def PayInvoice(req, ctx):
    """Try to pay a LN invoice from its payment request (bolt 11).

    An amount can be specified if the invoice doesn't already have it included.
    If a description hash is included in the invoice, its preimage must be
    included in the request.
    """
    check_req_params(ctx, req, 'payment_request')
    amount_encoded = has_amount_encoded(req.payment_request)
    res = pb.PayInvoiceResponse()
    lnd_req = ln.SendRequest(payment_request=req.payment_request)
    if req.cltv_expiry_delta:
        Enf.check_value(ctx,
                        req.cltv_expiry_delta,
                        enforce=Enf.CLTV_EXPIRY_DELTA)
        lnd_req.final_cltv_delta = req.cltv_expiry_delta
    if req.amount_msat and amount_encoded:
        Err().unsettable(ctx, 'amount_msat')
    elif req.amount_msat and not amount_encoded and Enf.check_value(
            ctx, req.amount_msat, LND_LN_TX):
        lnd_req.amt_msat = req.amount_msat
    elif not amount_encoded:
        check_req_params(ctx, req, 'amount_msat')
    with _connect(ctx) as stub:
        lnd_res = stub.SendPaymentSync(lnd_req, timeout=get_node_timeout(ctx))
        if lnd_res.payment_preimage:
            res.payment_preimage = hexlify(lnd_res.payment_preimage)
        elif lnd_res.payment_error:
            _handle_error(ctx, lnd_res.payment_error)
    return res


@_handle_rpc_errors
def PayOnChain(req, ctx):
    """Try to pay a bitcoin address."""
    check_req_params(ctx, req, 'address', 'amount_sat')
    Enf.check_value(ctx, req.amount_sat, enforce=Enf.OC_TX)
    lnd_req = ln.SendCoinsRequest(addr=req.address, amount=req.amount_sat)
    if req.fee_sat_byte:
        Enf.check_value(ctx, req.fee_sat_byte, enforce=Enf.OC_FEE)
        lnd_req.sat_per_byte = req.fee_sat_byte
    res = pb.PayOnChainResponse()
    with _connect(ctx) as stub:
        lnd_res = stub.SendCoins(lnd_req, timeout=get_node_timeout(ctx))
        res.txid = lnd_res.txid
    return res


@_handle_rpc_errors
def UnlockNode(req, ctx):
    """Try to unlock node."""
    return unlock_node_with_password(ctx, req, unlock_node)


# pylint: disable=too-many-arguments,too-many-branches
def _add_channel(res, lnd_chan, state, active_only=False, open_chan=False):
    """Add a channel to a ListChannelsResponse."""
    if active_only and not lnd_chan.active:
        return
    if lnd_chan.ListFields():
        if not open_chan:
            pending_chan = lnd_chan
            lnd_chan = pending_chan.channel
            if not lnd_chan.ListFields():
                return
        local_balance = lnd_chan.local_balance
        remote_balance = lnd_chan.remote_balance
        channel = res.channels.add(
            funding_txid=lnd_chan.channel_point.split(':')[0],
            capacity_msat=int(lnd_chan.capacity * 1000),
            state=state,
            local_reserve_sat=lnd_chan.local_chan_reserve_sat,
            remote_reserve_sat=lnd_chan.remote_chan_reserve_sat)
        if open_chan:
            if lnd_chan.initiator:
                if remote_balance < lnd_chan.local_constraints.dust_limit_sat:
                    local_balance += lnd_chan.commit_fee - remote_balance
                else:
                    local_balance += lnd_chan.commit_fee
            else:
                if local_balance < lnd_chan.local_constraints.dust_limit_sat:
                    remote_balance += lnd_chan.commit_fee - local_balance
                else:
                    remote_balance += lnd_chan.commit_fee
            channel.remote_pubkey = lnd_chan.remote_pubkey
            channel.channel_id = str(lnd_chan.chan_id)
            channel.to_self_delay = lnd_chan.csv_delay
            channel.private = lnd_chan.private
            channel.active = lnd_chan.active
        else:
            channel.remote_pubkey = lnd_chan.remote_node_pub
            channel.active = False
            commit_fee = 0
            if isinstance(pending_chan,
                          ln.PendingChannelsResponse.PendingOpenChannel):
                commit_fee = pending_chan.commit_fee
            if isinstance(pending_chan,
                          ln.PendingChannelsResponse.WaitingCloseChannel):
                commit_fee = lnd_chan.capacity - local_balance - remote_balance
            if commit_fee:
                if lnd_chan.initiator == ln.INITIATOR_LOCAL:
                    local_balance += commit_fee
                elif lnd_chan.initiator == ln.INITIATOR_REMOTE:
                    remote_balance += commit_fee
        if lnd_chan.capacity == local_balance + remote_balance:
            channel.local_balance_msat = int(local_balance * 1000)
            channel.remote_balance_msat = int(remote_balance * 1000)
    # pylint: enable=too-many-arguments,too-many-branches


def _add_invoice(res, lnd_invoice, invoice_state):
    """Add an invoice to a ListInvoicesResponse."""
    if lnd_invoice.ListFields():
        invoice = res.invoices.add(
            timestamp=lnd_invoice.creation_date,
            payment_hash=hexlify(lnd_invoice.r_hash),
            description=lnd_invoice.memo,
            description_hash=hexlify(lnd_invoice.description_hash),
            expiry=lnd_invoice.expiry,
            fallback_addr=lnd_invoice.fallback_addr,
            state=invoice_state,
            payment_request=lnd_invoice.payment_request,
            amount_received_msat=lnd_invoice.amt_paid_msat)
        invoice.amount_encoded_msat = lnd_invoice.value * 1000
        if lnd_invoice.value_msat:
            invoice.amount_encoded_msat = lnd_invoice.value_msat
        for lnd_route in lnd_invoice.route_hints:
            _add_route_hint(invoice, lnd_route)


def _add_payment(res, lnd_payment):
    """Add a payment to a ListPaymentsResponse."""
    if lnd_payment.ListFields():
        res.payments.add(payment_hash=lnd_payment.payment_hash,
                         amount_msat=lnd_payment.value_msat,
                         timestamp=int(lnd_payment.creation_time_ns / 10**9),
                         fee_msat=lnd_payment.fee_msat,
                         payment_preimage=lnd_payment.payment_preimage)


def _add_route_hint(res, lnd_route):
    """Add a route hint and its hop hints to a DecodeInvoiceResponse."""
    if lnd_route.ListFields():
        grpc_route = res.route_hints.add()
    for lnd_hop in lnd_route.hop_hints:
        grpc_route.hop_hints.add(
            pubkey=lnd_hop.node_id,
            short_channel_id=str(lnd_hop.chan_id),
            fee_base_msat=lnd_hop.fee_base_msat,
            fee_proportional_millionths=lnd_hop.fee_proportional_millionths,
            cltv_expiry_delta=lnd_hop.cltv_expiry_delta)


def _add_transaction(res, lnd_transaction):
    """Add a transaction to a ListTransactionsResponse."""
    if lnd_transaction.ListFields():
        res.transactions.add(txid=lnd_transaction.tx_hash,
                             amount_sat=lnd_transaction.amount,
                             confirmations=lnd_transaction.num_confirmations,
                             block_hash=lnd_transaction.block_hash,
                             block_height=lnd_transaction.block_height,
                             timestamp=lnd_transaction.time_stamp,
                             fee_sat=lnd_transaction.total_fees)


def _check_timestamp(req, lnd_invoice):
    """Decide if invoice has to be skipped (not added to response).

    Check creation_date against search_timestamp in order to decide.
    """
    if not req.search_timestamp:
        return False
    if req.list_order:
        # descending list_order
        if req.search_order:
            # descending search_order: use descending list, skip newer invoices
            # will stop when list reaches max_items size
            if lnd_invoice.creation_date >= req.search_timestamp:
                return True
        else:
            # ascending search_order: use ascending list, skip older invoices
            # must flip the list at the end
            if lnd_invoice.creation_date <= req.search_timestamp:
                return True
    else:
        # ascending list_order
        if req.search_order:
            # descending search_order: use descending list, skip newer invoices
            # must flip the list at the end
            if lnd_invoice.creation_date >= req.search_timestamp:
                return True
        else:
            # ascending search_order: use ascending list, skip older invoices
            # will stop when list reaches max_items size
            if lnd_invoice.creation_date <= req.search_timestamp:
                return True
    return False


@handle_thread
def _close_channel(lnd_req, close_timeout):
    """Close a LN channel and return a closing TXID or raise an exception."""
    lnd_res = None
    try:
        with _connect(FakeContext()) as stub:
            for lnd_res in stub.CloseChannel(lnd_req, timeout=close_timeout):
                LOGGER.debug('[ASYNC] CloseChannel released response: %s',
                             str(lnd_res).replace('\n', ''))
                if lnd_res.close_pending.txid:
                    lnd_res = _txid_bytes_to_str(lnd_res.close_pending.txid)
                    break
    except RpcError as err:
        # pylint: disable=no-member
        error = err.details() if hasattr(err, 'details') else err
        # pylint: enable=no-member
        LOGGER.debug('[ASYNC] CloseChannel terminated with error: %s', error)
        raise err
    except RuntimeError as err:
        raise err
    return lnd_res


@contextmanager
def _connect(ctx, stub_class=None, force_no_macaroon=False):
    """Securely connect to the lnd node using gRPC."""
    creds = sett.LND_CREDS_FULL
    if force_no_macaroon:
        creds = sett.LND_CREDS_SSL
    channel = secure_channel(sett.LND_ADDR, creds)
    future_channel = channel_ready_future(channel)
    try:
        future_channel.result(timeout=get_node_timeout(ctx))
    except FutureTimeoutError:
        # Handle gRPC channel that did not connect
        Err().node_error(ctx, 'Failed to dial server')
    else:
        if stub_class is None:
            stub_class = lnrpc.LightningStub
        stub = stub_class(channel)
        yield stub
        channel.close()


def _get_invoice_state(lnd_invoice):
    """Return the invoice state."""
    now = datetime.now().timestamp()
    if lnd_invoice.state == ln.Invoice.SETTLED:
        return pb.Invoice.PAID
    if lnd_invoice.state in (ln.Invoice.OPEN, ln.Invoice.ACCEPTED):
        if (lnd_invoice.creation_date + lnd_invoice.expiry) < now:
            return pb.Invoice.EXPIRED
        return pb.Invoice.PENDING
    if lnd_invoice.state == ln.Invoice.CANCELED:
        return pb.Invoice.EXPIRED
    return pb.Invoice.UNKNOWN


def _handle_error(ctx, error):
    """Report error received by lnd.

    This is always terminating: raises a grpc.RpcError to the API caller.
    """
    error = error.details() if hasattr(error, 'details') else error
    if not isinstance(error, str):
        error = 'Could not decode error message'
    Err().report_error(ctx, error)


def _metadata_callback(_ctx, callback):
    """Get lnd macaroon."""
    macaroon = encode(sett.LND_MAC, 'hex')
    callback([('macaroon', macaroon)], None)


def _parse_invoices(res, invoices, req):
    """Decide if invoice has to be added and which with state.

    Return True when added max_items invoices."""
    if not any([req.paid, req.pending, req.expired, req.unknown]):
        req.paid = req.pending = req.expired = req.unknown = True
    for lnd_invoice in invoices:
        if _check_timestamp(req, lnd_invoice):
            continue
        invoice_state = _get_invoice_state(lnd_invoice)
        if req.paid and invoice_state == pb.Invoice.PAID:
            _add_invoice(res, lnd_invoice, invoice_state)
        if req.pending and invoice_state == pb.Invoice.PENDING:
            _add_invoice(res, lnd_invoice, invoice_state)
        if req.expired and invoice_state == pb.Invoice.EXPIRED:
            _add_invoice(res, lnd_invoice, invoice_state)
        if req.unknown and invoice_state == pb.Invoice.UNKNOWN:
            _add_invoice(res, lnd_invoice, invoice_state)
        if len(res.invoices) == req.max_items:
            return True
    return False


def _txid_bytes_to_str(txid):
    """Decode big-endian TXID bytes to a little-endian TXID string."""
    return encode(txid[::-1], 'hex').decode()



# Boltlight - a LN node wrapper
#
# Copyright (C) 2021-2022 boltlight contributors
# Copyright (C) 2018 inbitcoin s.r.l.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# For a full list of contributors, please see the AUTHORS.md file.
"""Implementation of boltlight.proto defined methods for electrum."""

from ast import literal_eval
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import TimeoutError as TimeoutFutError
from contextlib import ExitStack, suppress
from json import dumps
from logging import getLogger

from . import boltlight_pb2 as pb
from . import settings as sett
from .errors import Err
from .utils.bitcoin import Enforcer as Enf
from .utils.bitcoin import (
    convert, get_address_type, get_channel_balances, has_amount_encoded)
from .utils.db import session_scope
from .utils.misc import handle_thread, set_defaults
from .utils.network import (
    FakeContext, JSONRPCSession, check_req_params, get_thread_timeout)
from .utils.security import get_secret, unlock_node_with_password

LOGGER = getLogger(__name__)

ERRORS = {
    'Bad bech32 checksum': {
        'fun': 'invalid',
        'params': 'payment_request'
    },
    'Forbidden': {
        'fun': 'wrong_node_password'
    },
    'Hostname does not resolve': {
        'fun': 'connect_failed'
    },
    'Invalid node ID, must be 33 bytes and hexadecimal': {
        'fun': 'invalid',
        'params': 'node_uri'
    },
    'This invoice has expired': {
        'fun': 'invoice_expired'
    },
    'wallet not loaded': {
        'fun': 'node_locked'
    }
}

ELE_LN_TX = {'min_value': 1, 'max_value': 2**32, 'unit': Enf.SATS}


def get_node_version():
    """Get node's version."""
    rpc_ele = ElectrumRPC()
    with suppress(RuntimeError):
        ele_res, is_err = rpc_ele.getinfo(FakeContext())
        if not is_err:
            return ele_res.get('version', '')
    return ''


def get_settings(config, sec):
    """Get electrum settings."""
    ele_values = ['ELE_HOST', 'ELE_PORT', 'ELE_USER']
    set_defaults(config, ele_values)
    sett.ELE_HOST = config.get(sec, 'ELE_HOST')
    sett.ELE_PORT = config.get(sec, 'ELE_PORT')
    sett.ELE_USER = config.get(sec, 'ELE_USER')
    sett.IMPL_SEC_TYPE = 'password'


def unlock_node(ctx, password, session=None):
    """Unlock node with password saved in boltlight's DB."""
    with ExitStack() if session else session_scope(ctx) as ses:
        if session:
            ses = session
        ele_pass = get_secret(ctx, ses, password, 'electrum', 'password')
        # update password, allowing to change it during boltlight execution
        update_settings(ele_pass)
        rpc_ele = ElectrumRPC()
        ele_res, is_err = rpc_ele.load_wallet(ctx)
        if is_err:
            _handle_error(ctx, ele_res)


def update_settings(password):
    """Update electrum specific settings."""
    ele_pass = password.decode()
    sett.RPC_URL = \
        f'http://{sett.ELE_USER}:{ele_pass}@{sett.ELE_HOST}:{sett.ELE_PORT}'


def BalanceOffChain(_req, ctx):
    """Return the off-chain balance available across all channels."""
    channels = ListChannels(pb.ListChannelsRequest(), ctx).channels
    return get_channel_balances(channels)


def BalanceOnChain(_req, ctx):
    """Return the on-chain balance in satoshi of the running LN node."""
    rpc_ele = ElectrumRPC()
    ele_res, is_err = rpc_ele.getbalance(ctx)
    if is_err:
        _handle_error(ctx, ele_res)
    res = pb.BalanceOnChainResponse(confirmed_sat=convert(
        ctx, Enf.BTC, Enf.SATS, ele_res.get('confirmed'), Enf.SATS))
    res.total_sat = res.confirmed_sat + convert(
        ctx, Enf.BTC, Enf.SATS, ele_res.get('unconfirmed'), Enf.SATS)
    return res


def CheckInvoice(req, ctx):
    """Check if a LN invoice has been paid."""
    check_req_params(ctx, req, 'payment_hash')
    rpc_ele = ElectrumRPC()
    ele_req = {'key': req.payment_hash}
    ele_res, is_err = rpc_ele.getrequest(ctx, ele_req)
    if is_err:
        _handle_error(ctx, ele_res)
    return pb.CheckInvoiceResponse(state=_get_invoice_state(ele_res))


def CloseChannel(req, ctx):
    """Try to close a LN channel."""
    check_req_params(ctx, req, 'channel_id')
    rpc_ele = ElectrumRPC()
    ele_res, is_err = rpc_ele.list_channels(ctx)
    if is_err:
        _handle_error(ctx, ele_res)
    ele_req = {}
    for ele_chan in ele_res:
        if ele_chan.get('channel_id') == req.channel_id:
            ele_req['channel_point'] = ele_chan['channel_point']
            break
    if 'channel_point' not in ele_req:
        Err().invalid(ctx, 'channel_id')
    if req.force:
        ele_req['force'] = True
    with ThreadPoolExecutor(max_workers=1) as executor:
        future = executor.submit(_close_channel, ele_req)
        try:
            ele_res = future.result(timeout=get_thread_timeout(ctx))
            if ele_res:
                return pb.CloseChannelResponse(closing_txid=ele_res)
        except RuntimeError as ele_err:
            try:
                error = literal_eval(str(ele_err))
                if isinstance(error, bytes):
                    error = ('Bytes received instead of string (hint: trying '
                             'to close a channel while the peer is offline?)')
                _handle_error(ctx, error)
            except (SyntaxError, ValueError):
                Err().report_error(ctx, str(ele_err))
        except TimeoutFutError:
            executor.shutdown(wait=False)
    return pb.CloseChannelResponse()


def CreateInvoice(req, ctx):
    """Create a LN invoice (bolt 11)."""
    check_req_params(ctx, req, 'amount_msat')
    if req.min_final_cltv_expiry:
        Err().unimplemented_parameter(ctx, 'min_final_cltv_expiry')
    if req.fallback_addr:
        Err().unimplemented_parameter(ctx, 'fallback_addr')
    amount_btc = convert(ctx,
                         Enf.MSATS,
                         Enf.BTC,
                         req.amount_msat,
                         Enf.MSATS,
                         enforce=ELE_LN_TX)
    ele_req = {'amount': amount_btc, 'expiration': sett.EXPIRY_TIME}
    if req.expiry:
        ele_req['expiration'] = req.expiry
    if req.description:
        ele_req['memo'] = req.description
    rpc_ele = ElectrumRPC()
    ele_res, is_err = rpc_ele.add_lightning_request(ctx, ele_req)
    if is_err:
        _handle_error(ctx, ele_res)
    timestamp = ele_res.get('timestamp')
    expires_in = ele_res.get('expiration')
    return pb.CreateInvoiceResponse(
        payment_request=ele_res.get('invoice'),
        payment_hash=ele_res.get('rhash'),
        expires_at=timestamp + expires_in if timestamp and expires_in else 0)


def DecodeInvoice(req, ctx):
    """Return information of an invoice from its payment request (bolt 11)."""
    check_req_params(ctx, req, 'payment_request')
    if req.description:
        Err().unimplemented_parameter(ctx, 'description')
    ele_req = {'invoice': req.payment_request}
    rpc_ele = ElectrumRPC()
    ele_res, is_err = rpc_ele.decode_invoice(ctx, ele_req)
    if is_err:
        _handle_error(ctx, ele_res)
    return pb.DecodeInvoiceResponse(amount_msat=int(
        ele_res.get('amount_msat', 0)),
                                    timestamp=ele_res.get('time'),
                                    payment_hash=ele_res.get('rhash'),
                                    destination_pubkey=ele_res.get('pubkey'),
                                    description=ele_res.get('message'),
                                    expiry=ele_res.get('exp'))


def GetNodeInfo(_req, ctx):
    """Return info about the running LN node."""
    rpc_ele = ElectrumRPC()
    ele_res, is_err = rpc_ele.nodeid(ctx)
    if is_err:
        _handle_error(ctx, ele_res)
    identity_pubkey = node_uri = ''
    if ele_res:
        identity_pubkey = ele_res
        if '@' in ele_res:
            identity_pubkey = ele_res.split('@')[0]
            node_uri = ele_res
    ele_res, is_err = rpc_ele.getinfo(ctx)
    if is_err:
        _handle_error(ctx, ele_res)
    res = pb.GetNodeInfoResponse(identity_pubkey=identity_pubkey,
                                 node_uri=node_uri,
                                 block_height=max(
                                     0, ele_res.get('blockchain_height')))
    ele_path = ele_res.get('path')
    if ele_path:
        res.network = pb.Network.MAINNET
        if ele_path.split('/')[-1] == 'testnet':
            res.network = pb.Network.TESTNET
        elif ele_path.split('/')[-1] == 'regtest':
            res.network = pb.Network.REGTEST
    return res


def ListChannels(req, ctx):
    """Return a list of channels of the running LN node."""
    rpc_ele = ElectrumRPC()
    ele_res, is_err = rpc_ele.list_channels(ctx)
    if is_err:
        _handle_error(ctx, ele_res)
    res = pb.ListChannelsResponse()
    for ele_chan in ele_res:
        _add_channel(ctx, res, ele_chan, req.active_only)
    return res


# pylint: disable=too-many-branches
def ListInvoices(req, ctx):
    """Return a list of lightning invoices created by the running LN node."""
    if not req.max_items:
        req.max_items = sett.MAX_INVOICES
    res = pb.ListInvoicesResponse()
    ele_req = {
        'paid': req.paid,
        'pending': req.pending,
        'expired': req.expired
    }
    filter_inv = True
    # Return unfiltered invoices list if no filter parameter is given
    if not any(
            getattr(req, f)
            for f in ('paid', 'pending', 'expired', 'unknown')):
        filter_inv = False
    rpc_ele = ElectrumRPC()
    ele_res, is_err = rpc_ele.list_requests(ctx, ele_req)
    if is_err:
        _handle_error(ctx, ele_res)
    ele_res = sorted(ele_res,
                     reverse=req.search_order,
                     key=lambda t: t.get('timestamp', 0))
    for ele_inv in ele_res:
        if req.search_timestamp:
            inv_ts = ele_inv.get('timestamp', 0)
            if req.search_order and inv_ts > req.search_timestamp:
                continue
            if not req.search_order and inv_ts < req.search_timestamp:
                continue
        state = _get_invoice_state(ele_inv)
        if filter_inv:
            if not req.paid and state == pb.Invoice.PAID:
                continue
            if not req.pending and state == pb.Invoice.PENDING:
                continue
            if not req.expired and state == pb.Invoice.EXPIRED:
                continue
            if not req.unknown and state == pb.Invoice.UNKNOWN:
                continue
        _add_invoice(res, ele_inv)
        if len(res.invoices) == req.max_items:
            break
    if req.list_order != req.search_order:
        res.CopyFrom(pb.ListInvoicesResponse(invoices=reversed(res.invoices)))
    return res


# pylint: enable=too-many-branches


def ListPayments(_req, ctx):
    """Return a list of lightning invoices paid by the running LN node."""
    rpc_ele = ElectrumRPC()
    ele_res, is_err = rpc_ele.lightning_history(ctx)
    if is_err:
        _handle_error(ctx, ele_res)
    res = pb.ListPaymentsResponse()
    for ele_payment in ele_res:
        _add_payment(res, ele_payment)
    return res


def ListPeers(_req, ctx):
    """Return a list of peers connected to the running LN node."""
    rpc_ele = ElectrumRPC()
    ele_res, is_err = rpc_ele.list_peers(ctx)
    if is_err:
        _handle_error(ctx, ele_res)
    res = pb.ListPeersResponse()
    for ele_peer in ele_res:
        res.peers.add(address=ele_peer.get('address'),
                      pubkey=ele_peer.get('node_id'))
        # alias and RGB color not available in electrum response
    return res


def ListTransactions(_req, ctx):
    """Return a list of on-chain transactions of the running LN node."""
    rpc_ele = ElectrumRPC()
    ele_res, is_err = rpc_ele.onchain_history(ctx)
    if is_err:
        _handle_error(ctx, ele_res)
    res = pb.ListTransactionsResponse()
    for ele_tx in ele_res.get('transactions', []):
        _add_transaction(ctx, res, ele_tx)
    return res


def NewAddress(req, ctx):
    """Create a new bitcoin address under control of the running LN node."""
    rpc_ele = ElectrumRPC()
    res = pb.NewAddressResponse()
    ele_req = {'unused': True, 'receiving': True}
    ele_res, is_err = rpc_ele.listaddresses(ctx, ele_req)
    if is_err:
        _handle_error(ctx, ele_res)
    LOGGER.debug('Listaddress response: %s, %s', ele_res, is_err)
    if ele_res:
        addr_type = get_address_type(ele_res[0])
        LOGGER.debug('The address types are: %s, %s', addr_type, req.addr_type)
        if addr_type != req.addr_type:
            Err().unimplemented_param_value(
                ctx, 'addr_type', pb.Address.Type.Name(req.addr_type))
        for addr in ele_res:
            if addr not in sett.ELE_RELEASED_ADDRESSES:
                sett.ELE_RELEASED_ADDRESSES.append(addr)
                res.address = addr
                break
    if not res.address:
        LOGGER.debug('The list of addresses provided by the listaddresses '
                     'API is exhausted, re-using existing ones')
        sett.ELE_RELEASED_ADDRESSES = [ele_res[0]]
        res.address = ele_res[0]
    return res


def OpenChannel(req, ctx):
    """Try to connect and open a channel with a peer."""
    check_req_params(ctx, req, 'node_uri', 'funding_sat')
    if not req.private:
        Err().unimplemented_param_value(ctx, 'private', 'False')
    rpc_ele = ElectrumRPC()
    amount_btc = convert(ctx,
                         Enf.SATS,
                         Enf.BTC,
                         req.funding_sat,
                         Enf.SATS,
                         enforce=Enf.FUNDING_SATOSHIS)
    ele_req = {'connection_string': req.node_uri, 'amount': amount_btc}
    if req.push_msat:
        if req.push_msat >= 1000 * req.funding_sat:
            Err().value_too_high(ctx, req.push_msat)
        ele_req['push_amount'] = convert(ctx,
                                         Enf.MSATS,
                                         Enf.BTC,
                                         req.push_msat,
                                         Enf.SATS,
                                         enforce=Enf.PUSH_MSAT)
    ele_res, is_err = rpc_ele.open_channel(ctx, ele_req)
    if is_err:
        _handle_error(ctx, ele_res)
    return pb.OpenChannelResponse(funding_txid=ele_res.split(':')[0])


def PayInvoice(req, ctx):
    """Try to pay a LN invoice from its payment request (bolt 11).

    Electrum doesn't currently support the payment of invoices:
    - with amount not set (or 0)
    - with description_hash encoded (description needed to decode/pay invoice)
    - that set a custom expiry (cltv_expiry_delta) for the payment
    """
    check_req_params(ctx, req, 'payment_request')
    amount_encoded = has_amount_encoded(req.payment_request)
    if amount_encoded and req.amount_msat:
        Err().unsettable(ctx, 'amount_msat')
    if req.amount_msat:
        Err().unimplemented_parameter(ctx, 'amount_msat')
    if not amount_encoded:
        Err().amount_required(ctx)
    if req.description:
        Err().unimplemented_parameter(ctx, 'description')
    if req.cltv_expiry_delta:
        Err().unimplemented_parameter(ctx, 'cltv_expiry_delta')
    rpc_ele = ElectrumRPC()
    ele_req = {'invoice': req.payment_request}
    ele_res, is_err = rpc_ele.lnpay(ctx, ele_req)
    if is_err:
        _handle_error(ctx, ele_res)
    elif not ele_res.get('success'):
        Err().payinvoice_failed(ctx)
    return pb.PayInvoiceResponse(payment_preimage=ele_res.get('preimage'))


def PayOnChain(req, ctx):
    """Try to pay a bitcoin address."""
    check_req_params(ctx, req, 'address', 'amount_sat')
    Enf.check_value(ctx, req.amount_sat, enforce=Enf.OC_TX)
    ele_req = {'destination': req.address, 'amount': req.amount_sat}
    if req.fee_sat_byte:
        Enf.check_value(ctx, req.fee_sat_byte, enforce=Enf.OC_FEE)
        ele_req['feerate'] = req.fee_sat_byte
    rpc_ele = ElectrumRPC()
    ele_res, is_err = rpc_ele.payto(ctx, ele_req)
    if is_err:
        _handle_error(ctx, ele_res)
    ele_req = {'tx': ele_res}
    ele_res, is_err = rpc_ele.broadcast(ctx, ele_req)
    if is_err:
        _handle_error(ctx, ele_res)
    return pb.PayOnChainResponse(txid=ele_res)


def UnlockNode(req, ctx):
    """Try to unlock node."""
    return unlock_node_with_password(ctx, req, unlock_node)


def _add_channel(ctx, res, ele_chan, active_only):
    """Add a channel to a ListChannelsResponse."""
    state = _get_channel_state(ele_chan)
    active = state == pb.Channel.OPEN and ele_chan.get('peer_state') == 'GOOD'
    if state < 0 or (active_only and not active):
        return
    grpc_chan = res.channels.add(
        state=state,
        remote_pubkey=ele_chan.get('remote_pubkey'),
        short_channel_id=ele_chan.get('short_channel_id'),
        channel_id=ele_chan.get('channel_id'),
        active=active,
        local_balance_msat=convert(ctx, Enf.SATS, Enf.MSATS,
                                   ele_chan.get('local_balance'), Enf.MSATS),
        remote_balance_msat=convert(ctx, Enf.SATS, Enf.MSATS,
                                    ele_chan.get('remote_balance'), Enf.MSATS))
    loc_reserve = ele_chan.get('local_reserve', 0)
    grpc_chan.local_reserve_sat = int(loc_reserve) if loc_reserve else 0
    rem_reserve = ele_chan.get('remote_reserve', 0)
    grpc_chan.remote_reserve_sat = int(rem_reserve) if rem_reserve else 0
    chan_point = ele_chan.get('channel_point')
    if chan_point:
        grpc_chan.funding_txid = chan_point.split(':')[0]
    grpc_chan.capacity_msat = \
        grpc_chan.remote_balance_msat + grpc_chan.local_balance_msat


def _add_invoice(res, ele_inv):
    """Add an invoice to a ListInvoicesResponse."""
    invoice = res.invoices.add(amount_encoded_msat=ele_inv.get('amount_msat'),
                               timestamp=ele_inv.get('timestamp'),
                               payment_hash=ele_inv.get('rhash'),
                               description=ele_inv.get('message'),
                               state=_get_invoice_state(ele_inv),
                               payment_request=ele_inv.get('invoice'))
    expires_in = ele_inv.get('expiration')
    invoice.expiry = invoice.timestamp + expires_in \
        if invoice.timestamp and expires_in else 0


def _add_payment(res, ele_payment):
    """Add a payment to a ListPaymentsResponse."""
    if (ele_payment.get('type') == 'payment'
            and ele_payment.get('direction') == 'sent'):
        res.payments.add(payment_hash=ele_payment.get('payment_hash'),
                         amount_msat=int(-ele_payment.get('amount_msat', 0)),
                         timestamp=ele_payment.get('timestamp'),
                         fee_msat=ele_payment.get('fee_msat'),
                         payment_preimage=ele_payment.get('preimage'))


def _add_transaction(ctx, res, ele_tx):
    """Add a transaction to a ListTransactionsResponse."""
    res.transactions.add(txid=ele_tx.get('txid'),
                         confirmations=ele_tx.get('confirmations'),
                         block_height=ele_tx.get('height'),
                         timestamp=ele_tx.get('timestamp'),
                         fee_sat=ele_tx.get('fee_sat'),
                         amount_sat=convert(ctx, Enf.BTC, Enf.SATS,
                                            ele_tx.get('bc_value'), Enf.SATS))


@handle_thread
def _close_channel(ele_req):
    """Close a LN channel and return a closing TXID or raise an exception."""
    rpc_ele = ElectrumRPC()
    ele_res = error = None
    try:
        ele_res, is_err = rpc_ele.close_channel(FakeContext(), ele_req)
        if is_err:
            error = ele_res
        else:
            LOGGER.debug('[ASYNC] CloseChannel terminated with response: %s',
                         ele_res)
    except RuntimeError as err:
        error = str(err)
    if error:
        LOGGER.debug('[ASYNC] CloseChannel terminated with error: %s', error)
        raise RuntimeError(error)
    return ele_res


def _get_channel_state(ele_chan):
    """Return the channel state."""
    state = ele_chan.get('state')
    if state in (
            'CLOSED',
            'REDEEMED',
    ):
        return -1
    if state in (
            'FUNDED',
            'OPENING',
            'PREOPENING',
    ):
        return pb.Channel.PENDING_OPEN
    if state in ('OPEN', ):
        return pb.Channel.OPEN
    if state in ('FORCE_CLOSING', ):
        return pb.Channel.PENDING_FORCE_CLOSE
    if state in ('CLOSING', 'SHUTDOWN'):
        return pb.Channel.PENDING_MUTUAL_CLOSE
    return pb.Channel.UNKNOWN


def _get_invoice_state(ele_inv):
    """Return the invoice state.

    States of electrum payment requests:
    PR_UNPAID   = 0
    PR_EXPIRED  = 1
    PR_UNKNOWN  = 2     # sent but not propagated
    PR_PAID     = 3     # send and propagated
    PR_INFLIGHT = 4     # unconfirmed
    PR_FAILED   = 5
    PR_ROUTING  = 6
    """
    status = ele_inv.get('status')
    if status in (3, ):
        return pb.Invoice.PAID
    if status in (
            0,
            2,
            4,
            6,
    ):
        return pb.Invoice.PENDING
    if status in (1, ):
        return pb.Invoice.EXPIRED
    return pb.Invoice.UNKNOWN


def _handle_error(ctx, ele_res):
    """Report errors of an electrum RPC response.

    This is always terminating: raises a grpc.RpcError to the API caller.
    """
    Err().report_error(ctx, ele_res)


class ElectrumRPC(JSONRPCSession):
    """Create and mantain an RPC session with electrum."""
    def __init__(self):
        super().__init__(headers={'content-type': 'application/json'})

    def __getattr__(self, name):
        def call_adapter(ctx, params=None, timeout=None):
            if not params:
                params = {}
            payload = dumps({
                'id': self._id_count,
                'method': name,
                'params': params,
                'jsonrpc': self._jsonrpc_ver
            })
            LOGGER.debug('RPC req: %s', payload)
            # pylint: disable=super-with-arguments
            return super(ElectrumRPC, self).call(ctx, payload, timeout=timeout)

        return call_adapter
